"""
Base class for the instrument drivers for the Rohde&Schwarz Signal Generators
"""

import logging
import re
import time
from typing import List, Optional

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException, QMI_TimeoutException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.core.transport import create_transport

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class RohdeSchwarz_Base(QMI_Instrument):
    """Base class for the instrument driver for the Rohde&Schwarz Signal Generators."""

    # Default response timeout in seconds.
    DEFAULT_RESPONSE_TIMEOUT = 5.0

    def __init__(self,
                 context: QMI_Context,
                 name: str,
                 transport: str,
                 max_continuous_power: Optional[float] = None
                 ) -> None:
        """Initialize the instrument driver.

        Parameters:
            name:                   Name for this instrument instance.
            transport:              QMI transport descriptor to connect to the instrument.
            max_continuous_power:   Maximum allowable continuous RF power. The RF output power can be set higher than
                                    this value only when pulse modulation is enabled. Set to `None` to disable the
                                    limit (allow any output power).
        """
        super().__init__(context, name)
        self._timeout = self.DEFAULT_RESPONSE_TIMEOUT
        self._calibrating = False
        self._calibration_result = None  # type: Optional[int]
        self._calibration_error = None  # type: Optional[str]
        self._max_continuous_power = max_continuous_power
        self._transport = create_transport(transport, default_attributes={"port": 5025})
        self._scpi_protocol = ScpiProtocol(self._transport,
                                           command_terminator="\r\n",
                                           response_terminator="\n",
                                           default_timeout=self._timeout)

    def _is_valid_param(self, inp: str, params: List[str]) -> str:
        """ Checks if input string is a valid parameter string.

        Parameters:
            inp:    String parameter to be checked.
            params: Allowed parameter strings as a list

        Raises:
            ValueError: If the input value is not in the allowed parameters list.

        Return:
            The input string in upper case.
        """
        if inp.upper() in params:
            return inp.upper()
        else:
            raise ValueError('Values that can be set are %s' % ','.join(params))

    @rpc_method
    def open(self) -> None:
        """Open connection to the instrument."""
        _logger.info("[%s] Opening connection to instrument", self._name)
        self._check_is_closed()
        self._transport.open()
        self._calibrating = False
        self._calibration_result = None
        self._calibration_error = None
        super().open()

    @rpc_method
    def close(self) -> None:
        """Close connection to the instrument."""
        _logger.info("[%s] Closing connection to instrument", self._name)
        super().close()
        self._transport.close()

    def _ask_float(self, cmd: str) -> float:
        """Send a query and return a floating point response."""
        resp = self._scpi_protocol.ask(cmd)
        try:
            return float(resp)
        except ValueError as exc:
            raise QMI_InstrumentException(f"Unexpected response to command {cmd!r}: {resp!r}") from exc

    def _ask_bool(self, cmd: str) -> bool:
        """Send a query and return a boolean response."""
        resp = self._scpi_protocol.ask(cmd)
        value = resp.strip().upper()
        if value in ("1", "ON"):
            return True
        elif value in ("0", "OFF"):
            return False
        else:
            raise QMI_InstrumentException(f"Unexpected response to command {cmd!r}: {resp!r}")

    def _check_error(self) -> None:
        """Read the instrument error queue and raise an exception if there is an error."""
        resp = self._scpi_protocol.ask("SYST:ERR:ALL?")
        # When there are no errors, the response is '0,"No error"'.
        if not re.match(r"^\s*0\s*,", resp):
            # Some error occurred.
            raise QMI_InstrumentException(f"Instrument returned error: {resp}")

    def _internal_poll_calibration_finished(self) -> bool:
        """Check for ongoing calibration.

        Attributes:
            self._calibrating: Set to `False` if response is received from instrument.
            self._calibration_result: Integer result from the response or `None`.
            self._calibration_error:  String describing the erroneous response or `None`.

        Return:
            True if the calibration is finished (with or without error), else False.
        """
        assert self._calibrating
        assert self._transport is not None

        # Poll to see if the instrument sent a response to the calibration command.
        try:
            resp = self._transport.read_until(message_terminator=b"\n", timeout=0.001)
        except QMI_TimeoutException:
            # No response received; calibration still not finished.
            return False

        # Got response to calibration query.
        # Store it for future reference.
        self._calibrating = False
        self._calibration_result = None
        self._calibration_error = None
        resp = resp.rstrip(b"\r\n")
        try:
            self._calibration_result = int(resp)
        except ValueError:
            _logger.warning("Unexpected response to calibration command: %r", resp)
            self._calibration_error = f"Unexpected response to calibration command: {resp!r}"
            return True

        _logger.info("Internal adjustments finished with status %d", self._calibration_result)

        # Read error queue to see if calibration produced an error.
        calibration_resp = self._scpi_protocol.ask("SYST:ERR:ALL?")

        # When there are no errors, the response is '0,"No error"'.
        if not re.match(r"^\s*0\s*,", calibration_resp):
            # Some error occurred.
            _logger.warning("Calibration failed with error: %s", calibration_resp)
            self._calibration_error = f"Instrument returned error: {calibration_resp}"

        # Calibration finished.
        return True

    def _check_calibrating(self) -> None:
        """Check if a calibration is still running.

        Raises:
            QMI_InstrumentException: If the calibration is still in progress.
        """
        self._check_is_open()
        if self._calibrating:
            # Calibration started and result not yet received.
            # Poll to see if there is a result now.
            if not self._internal_poll_calibration_finished():
                # Calibration still not finished.
                raise QMI_InstrumentException("Calibration in progress") from None

            assert not self._calibrating

    def _get_external_reference_frequency(self) -> str:
        """Return the currently configured external reference input frequency.
        """
        self._check_calibrating()
        return self._scpi_protocol.ask(":ROSC:EXT:FREQ?").strip()

    def _set_external_reference_frequency(self, frequency: str, options: List[str]) -> None:
        """Configure the external reference input frequency.

        Parameters:
            frequency:  desired frequency.
            options:    allowed frequencies.
        """
        frequency = self._is_valid_param(frequency, options)
        self._check_calibrating()
        self._scpi_protocol.write(f":ROSC:EXT:FREQ {frequency}")
        self._check_error()

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return QMI_InstrumentIdentification instance.

        Raises:
            QMI_InstrumentException: on unexpected response to query.

        Returns:
            QMI_InstrumentIdentification: Instance of the instrument.
        """
        self._check_calibrating()
        resp = self._scpi_protocol.ask("*IDN?")
        words = resp.rstrip().split(",")
        if len(words) != 4:
            raise QMI_InstrumentException(f"Unexpected response to *IDN?, got {resp!r}")
        return QMI_InstrumentIdentification(vendor=words[0].strip(),
                                            model=words[1].strip(),
                                            serial=words[2].strip(),
                                            version=words[3].strip())

    @rpc_method
    def start_calibration(self) -> None:
        """Start internal adjustments. Needs to be implemented further in the deriving classes.

        This function returns immediately after starting the calibration.

        Calibration can take up to 10 minutes. Call poll_calibration() to see
        whether calibration is complete. No other commands can be processed
        while the instrument is calibrating.

        The instrument must be at stable temperature (30 minutes to warm up)
        before starting internal adjustments.

        Attributes:
            self._calibrating: Set to `True` if starting the calibration succeeded.

        Raises:
            QMI_InstrumentException: If the result of previous calibration is still pending
        """
        self._check_calibrating()
        if self._calibration_result is not None:
            raise QMI_InstrumentException("Result of previous calibration is still pending")

        _logger.info("Starting internal adjustments")
        self._calibrating = True

    @rpc_method
    def poll_calibration(self) -> Optional[int]:
        """Check whether an ongoing calibration is finished.

        Raises:
            QMI_InstrumentException on calibration error.

        Returns:
            If calibration is not yet finished, this function returns None.
            If calibration is finished, this function returns a status code:
                0 if calibration was successful;
                1 if calibration failed.
        """
        self._check_is_open()
        if self._calibrating:
            if not self._internal_poll_calibration_finished():
                # Still calibrating.
                return None
        assert not self._calibrating

        if self._calibration_error is not None:
            # Report calibration error.
            errmsg = self._calibration_error
            self._calibration_result = None
            self._calibration_error = None
            raise QMI_InstrumentException(errmsg)

        if self._calibration_result is not None:
            # Return calibration result.
            ret = self._calibration_result
            self._calibration_result = None
            return ret

        # No calibration active, no result, no error.
        raise QMI_InstrumentException("No calibration active")

    @rpc_method
    def reset(self) -> None:
        """Reset the instrument, returning (most) settings to their defaults."""
        # Unfortunately RST does not cancel ongoing calibration.
        self._check_calibrating()
        self._scpi_protocol.write("*CLS")  # clear error queue
        self._scpi_protocol.write("*RST")
        time.sleep(0.1)
        self._check_error()

    @rpc_method
    def get_phase(self) -> float:
        """Return the current phase delta in degrees."""
        self._check_calibrating()
        return self._ask_float(":PHAS?")

    @rpc_method
    def set_phase(self, phase: float) -> None:
        """Set the current phase delta.

        Parameters:
            phase:  target phase delta in degrees.
        """
        self._check_calibrating()
        self._scpi_protocol.write(f":PHAS {phase}")
        self._check_error()

    @rpc_method
    def get_reference_source(self) -> str:
        """Return the current reference source.

        Possible values:
          "INT" if the internal 10 MHz reference source is used;
          "EXT" if an external reference source is used.
        """
        self._check_calibrating()
        return self._scpi_protocol.ask(":ROSC:SOUR?").strip().upper()

    @rpc_method
    def set_reference_source(self, source: str) -> None:
        """Set the reference source.

        Parameters:
            source: desired reference source (accepted values: "INT", "EXT"); see also get_reference_source().
        """
        options = ["INT", "EXT"]
        source = self._is_valid_param(source, options)
        self._check_calibrating()
        self._scpi_protocol.write(f":ROSC:SOUR {source}")
        self._check_error()

    @rpc_method
    def get_frequency(self) -> float:
        """Return the current RF frequency in Hz."""
        self._check_calibrating()
        return self._ask_float(":FREQ?")

    @rpc_method
    def set_frequency(self, frequency: float) -> None:
        """Set the RF frequency.

        Parameters:
            frequency:  target frequency in Hertz.
        """
        self._check_calibrating()
        self._scpi_protocol.write(f":FREQ {frequency}")
        self._check_error()

    @rpc_method
    def get_pulsemod_ext_source(self) -> bool:
        """Check pulse modulation source.

        Return:
            True if pulse modulation uses an external source, else False.

        Raises:
            QMI_InstrumentException: On unexpected response.
        """
        self._check_calibrating()
        source = self._scpi_protocol.ask(":PULM:SOUR?").strip().upper()
        if source.startswith("EXT"):
            return True
        elif source.startswith("INT"):
            return False

        raise QMI_InstrumentException(f"Unexpected response {source!r} to command :PULM:SOUR?")

    @rpc_method
    def set_pulsemod_ext_source(self, ext: bool) -> None:
        """Enable or disable the external pulse modulation source.

        When set to True, pulse modulation is controlled by an external
        pulse generator connected to the TRIG port of the instrument.
        When set to False, pulse modulation is controlled by an internal
        pulse generator.

        The nominal threshold of the TRIG input signal is +1 Volt.
        The signal on the TRIG input must not exceed 5 Volt.

        Parameters:
            ext:    Boolean flag indicating if external modulation must be enabled.

        Raises:
            QMI_InstrumentException: If using internal source and power above max power limit.
        """
        source = "EXT" if ext else "INT"
        self._check_calibrating()

        # Safety check to prevent sample overheating.
        if (self._max_continuous_power is not None) and (not ext):
            if self.get_power() > self._max_continuous_power:
                raise QMI_InstrumentException(
                    f"Power limited to {self._max_continuous_power} dBm " +
                    "unless external pulse modulation source is selected"
                )

        self._scpi_protocol.write(f":PULM:SOUR {source}")
        self._check_error()

    @rpc_method
    def get_power_on_output_policy(self) -> str:
        """Return instrument power-on policy for setting the RF output state.

        Possible values:
          "OFF" if the instrument starts with RF output disabled;
          "UNCH" if the instrument starts with RF output state the same as before power down.
        """
        self._check_calibrating()
        return self._scpi_protocol.ask(":OUTP:PON?").strip().upper()

    @rpc_method
    def set_power_on_output_policy(self, policy: str) -> None:
        """Set the instrument power-on policy for setting the RF output state.

        Parameters:
            policy: desired policy (accepted values: "OFF", "UNCH"); see also get_power_on_output_policy().
        """
        options = ["OFF", "UNCH"]
        policy = self._is_valid_param(policy, options)
        self._check_calibrating()
        self._scpi_protocol.write(f":OUTP:PON {policy}")
        self._check_error()

    @rpc_method
    def get_pulsemod_polarity(self) -> bool:
        """Get pulse modulation polarity.

        Raises:
            QMI_InstrumentException: On unexpected response.

        Return:
             True if the external pulse modulation polarity is inverted; False if it is normal.
        """
        self._check_calibrating()
        resp = self._scpi_protocol.ask(":PULM:POL?").strip().upper()
        if resp.startswith("INV"):
            return True
        elif resp.startswith("NORM"):
            return False
        raise QMI_InstrumentException(f"Unexpected response {resp!r} ot command :PULM:POL?")

    @rpc_method
    def set_pulsemod_polarity(self, inverted: bool) -> None:
        """Enable or disable inverted polarity for external modulation.

        Parameters:
            inverted:   Boolean flag indicating if external modulation should be inverted.
                        For normal polarity (False), the RF signal is suppressed when the TRIG pulse is low.
                        For inverted polarity (True), the RF signal is suppressed when the TRIG pulse is high.

        Raises:
            QMI_InstrumentException: On using power beyond the max power limit.
        """
        self._check_calibrating()

        # Safety check to prevent sample overheating.
        if (self._max_continuous_power is not None) and inverted:
            if self.get_power() > self._max_continuous_power:
                raise QMI_InstrumentException(
                    "Power limited to {} dBm unless pulse modulation source is non-inverted".format(
                        self._max_continuous_power))

        self._scpi_protocol.write(f":PULM:POL {'INV' if inverted else 'NORM'}")
        self._check_error()

    @rpc_method
    def get_power(self) -> float:
        """Return the current RF output power in dBm."""
        self._check_calibrating()
        return self._ask_float(":POW?")

    @rpc_method
    def set_power(self, power: float) -> None:
        """Set the RF output power in dBm.

        Parameters:
            power:  target output power in dBm.

        Raises:
            QMI_InstrumentException: On power exceeding the power limit.
        """
        self._check_calibrating()

        # Safety check to prevent sample overheating.
        if (self._max_continuous_power is not None) and (power > self._max_continuous_power):
            if not self.get_pulsemod_enabled():
                raise QMI_InstrumentException(
                    "Power limited to {} dBm unless pulse modulation is enabled".format(self._max_continuous_power))
            if not self.get_pulsemod_ext_source():
                raise QMI_InstrumentException(
                    "Power limited to {} dBm unless external pulse modulation source is selected".format(
                        self._max_continuous_power))
            if self.get_pulsemod_polarity():
                raise QMI_InstrumentException(
                    "Power limited to {} dBm unless external pulse source polarity is non-inverted".format(
                        self._max_continuous_power))

        self._scpi_protocol.write(f":POW {power}")
        self._check_error()

    @rpc_method
    def get_output_state(self) -> bool:
        """Return True if RF output is enabled, False if RF output is disabled."""
        self._check_calibrating()
        return self._ask_bool(":OUTP?")

    @rpc_method
    def set_output_state(self, enable: bool) -> None:
        """Enable or disable RF output.

        Parameters:
            enable: target enabled state.

        Raises:
            QMI_InstrumentException: On power exceeding the power limit.
        """
        self._check_calibrating()

        # Safety check to prevent sample overheating.
        if (self._max_continuous_power is not None) and enable:
            if self.get_power() > self._max_continuous_power:
                if not self.get_pulsemod_enabled():
                    raise QMI_InstrumentException(
                        "Power limited to {} dBm unless pulse modulation is enabled".format(
                            self._max_continuous_power
                        ))
                if not self.get_pulsemod_ext_source():
                    raise QMI_InstrumentException(
                        "Power limited to {} dBm unless external pulse modulation source is selected".format(
                            self._max_continuous_power
                        ))
                if self.get_pulsemod_polarity():
                    raise QMI_InstrumentException(
                        "Power limited to {} dBm unless external pulse source polarity is non-inverted".format(
                            self._max_continuous_power
                        ))

        self._scpi_protocol.write(f":OUTP {1 if enable else 0}")
        self._check_error()

    @rpc_method
    def get_pulsemod_enabled(self) -> bool:
        """Return True if pulse modulation is enabled, False if disabled."""
        self._check_calibrating()
        return self._ask_bool(":PULM:STAT?")

    @rpc_method
    def set_pulsemod_enabled(self, enable: bool) -> None:
        """Enable or disable pulse modulation.

        Parameters:
            enable: target enabled state.

        Raises:
            QMI_InstrumentException: On power exceeding the power limit.
        """
        self._check_calibrating()

        # Safety check to prevent sample overheating.
        if (self._max_continuous_power is not None) and (not enable):
            if self.get_power() > self._max_continuous_power:
                raise QMI_InstrumentException(
                    "Pulse modulation cannot be enabled as power is set to {} dBm which is higher than the maximum {} dBm".format(
                        self.get_power(), self._max_continuous_power))

        self._scpi_protocol.write(f":PULM:STAT {1 if enable else 0}")
        self._check_error()

    @rpc_method
    def get_iq_enabled(self) -> bool:
        """Return True if IQ modulation is enabled, False if disabled."""
        self._check_calibrating()
        return self._ask_bool(":IQ:STAT?")

    @rpc_method
    def set_iq_enabled(self, enable: bool) -> None:
        """Enable or disable IQ modulation.

        Parameters:
            enable: target enabled state.
        """
        self._check_calibrating()
        self._scpi_protocol.write(f":IQ:STAT {1 if enable else 0}")
        self._check_error()

    @rpc_method
    def get_iq_wideband(self) -> bool:
        """Return True if wideband IQ modulation is enabled, False if disabled."""
        self._check_calibrating()
        return self._ask_bool(":IQ:WBST?")

    @rpc_method
    def set_iq_wideband(self, enable: bool) -> None:
        """Enable or disable wideband IQ modulation.

        Parameters:
            enable: target enabled state.
        """
        self._check_calibrating()
        self._scpi_protocol.write(f":IQ:WBST {1 if enable else 0}")
        self._check_error()

    @rpc_method
    def get_iq_quadrature_offset(self) -> float:
        """Return the current IQ quadrature offset."""
        self._check_calibrating()
        return self._ask_float(":IQ:IMP:QUAD?")

    @rpc_method
    def set_iq_quadrature_offset(self, phase: float) -> None:
        """Set the IQ quadrature offset between -8 and 8 degrees in increments of 0.01.

        Parameters:
            phase:  desired phase offset in degrees.

        Raises:
            ValueError: If phase offset not within -8 - 8 degrees.
        """
        if not -8.0 <= phase <= 8.0:
            raise ValueError("Phase offset should be in [-8, 8].")
        self._check_calibrating()
        self._scpi_protocol.write(f":IQ:IMP:QUAD {phase:.2f}")
        self._check_error()

    @rpc_method
    def get_iq_leakage_i(self) -> float:
        """Return the current I leakage amplitude (percent)."""
        self._check_calibrating()
        return self._ask_float(":IQ:IMP:LEAK:I?")

    @rpc_method
    def set_iq_leakage_i(self, leakage: float) -> None:
        """Set the I leakage amplitude between -5 and 5 (percent), in increments of 0.01.

        Parameters:
            leakage:    leakage amplitude in percent.

        Raises:
            ValueError: If leakage amplitude not within -5 - 5 percent.
        """
        if not -5.0 <= leakage <= 5.0:
            raise ValueError("Leakage offset should be in [-5, 5].")
        self._check_calibrating()
        self._scpi_protocol.write(f":IQ:IMP:LEAK:I {leakage:.2f}")
        self._check_error()

    @rpc_method
    def get_iq_leakage_q(self) -> float:
        """Return the current Q leakage amplitude (percent)."""
        self._check_calibrating()
        return self._ask_float(":IQ:IMP:LEAK:Q?")

    @rpc_method
    def set_iq_leakage_q(self, leakage: float) -> None:
        """Set the Q leakage amplitude between -5 and 5 (percent), in increments of 0.01.

        Parameters:
            leakage:    leakage amplitude in percent.

        Raises:
            ValueError: If leakage amplitude not within -5 - 5 percent.
        """
        if not -5.0 <= leakage <= 5.0:
            raise ValueError("Leakage offset should be in [-5, 5].")
        self._check_calibrating()
        self._scpi_protocol.write(f":IQ:IMP:LEAK:Q {leakage:.2f}")
        self._check_error()

    @rpc_method
    def get_iq_gain_imbalance(self) -> float:
        """Return the current IQ gain imbalance (dB)."""
        self._check_calibrating()
        return self._ask_float(":IQ:IMP:IQR:MAGN?")

    @rpc_method
    def set_iq_gain_imbalance(self, gain: float) -> None:
        """Set the IQ gain imbalance in dB in range -1 to 1, increments of 0.001.

        Parameters:
            gain:   desired gain in dB.

        Raises:
            ValueError: If gain imbalance not within -1 - 1 dB.
        """
        if not -1.0 <= gain <= 1.0:
            raise ValueError("Gain imbalance should be in [-1, 1].")
        self._check_calibrating()
        self._scpi_protocol.write(f":IQ:IMP:IQR:MAGN {gain:.3f}")
        self._check_error()

    @rpc_method
    def get_iq_crest_factor(self) -> float:
        """Return the current IQ crest factor compensation in dB."""
        self._check_calibrating()
        return self._ask_float(":IQ:CRES?")

    @rpc_method
    def set_iq_crest_factor(self, factor: float) -> None:
        """Set the IQ crest factor compensation in dB.

        Parameters:
            factor: crest factor in dB.
        """
        self._check_calibrating()
        self._scpi_protocol.write(f":IQ:CRES {factor}")
        self._check_error()

    @rpc_method
    def get_errors(self) -> None:
        """
        Queries the error/event queue for all unread items and
        removes them from the queue.
        """
        return self._check_error()

    @rpc_method
    def get_error_queue_length(self) -> int:
        """
        Queries the number of entries in the error queue.
        """
        return int(self._scpi_protocol.ask('SYST:ERR:COUN?'))


"""Implementation of the QMI_Transport class."""

import logging
import re
import socket
import sys
import time
if sys.version_info >= (3, 9):
    from collections.abc import Mapping
else:
    from typing import Mapping
from typing import Any, Dict, Optional, List, Tuple, Type

import serial
import vxi11  # type: ignore

from qmi.core.context import QMI_Context
from qmi.core.exceptions import (
    QMI_InvalidOperationException, QMI_TimeoutException, QMI_EndOfInputException,
    QMI_TransportDescriptorException, QMI_InstrumentException, QMI_RuntimeException
)
from qmi.core.util import format_address_and_port

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class QMI_Transport:
    """QMI_Transport is the base class for bidirectional byte stream transport implementations,
    typically used to talk to instruments.

    An instance of QMI_Transport represents a channel that admits reading and writing of arbitrary
    byte sequences. Message boundaries are not preserved. Subclasses of QMI_Transport implement the
    transport API for specific types of communication channels.

    Once created, a QMI_Transport needs to be opened via the open() method before reading and writing.
    When the application has finished using the transport, it must call the close() method to close
    the underlying channel and release system resources.
    """

    def __init__(self) -> None:
        """Initialize the transport.

        Subclasses must extend this method and set self._is_open to True
        when the transport is successfully initialized.
        """
        self._is_open = False

    def _check_is_open(self) -> None:
        """Verify that the transport is open, otherwise raise exception."""
        if not self._is_open:
            raise QMI_InvalidOperationException(
                "Operation not allowed on closed transport {}".format(type(self).__name__))

    def _open_transport(self) -> None:
        """Subclasses must override this method to open specific resources."""
        raise NotImplementedError("This is prototype function that has to be implemented by the inheriting sub-class")

    def open(self) -> None:
        """Open the transport and claim associated resources."""
        if self._is_open:
            raise QMI_InvalidOperationException(
                "Operation not allowed on opened transport {}".format(type(self).__name__))

        self._open_transport()
        self._is_open = True

    def close(self) -> None:
        """Close the transport and release associated resources.

        This transport instance can never be used again after it has been closed.

        Subclasses must override this method to close specific resources. When overriding this method, the subclass
        should make a call to this method before all resources are closed.
        """
        self._check_is_open()
        self._is_open = False

    def write(self, data: bytes) -> None:
        """Write a sequence of bytes to the transport.

        When this method returns, all bytes are written to the transport
        or queued to be written to the transport.

        An exception is raised if the transport is closed from the remote
        side before all bytes could be written.

        Subclasses must override this method, if applicable.
        """
        raise NotImplementedError("QMI_Transport.write not implemented")

    def read(self, nbytes: int, timeout: Optional[float]) -> bytes:
        """Read a specified number of bytes from the transport.

        This method blocks until the specified number of bytes are available,
        then returns the received bytes.

        If "timeout" is not None and the timeout expires before the requested
        number of bytes are available, QMI_TimeoutException is raised and
        any available bytes remain in the input buffer. If "timeout" is None,
        this method waits until the requested number of bytes are received.

        Subclasses must override this method, if applicable.

        Parameters:
            nbytes: Number of bytes to read.
            timeout: Maximum time to wait (in seconds), or None to wait indefinitely.

        Returns:
            Received bytes.

        Raises:
            ~qmi.core.exceptions.QMI_TimeoutException: If the timeout expires before the
                requested number of bytes are available.
            ~qmi.core.exceptions.QMI_EndOfInputException: If the transport has been closed on
                the remote side before the requested number of bytes are available.
        """
        raise NotImplementedError("QMI_Transport.read not implemented")

    def read_until(self, message_terminator: bytes, timeout: Optional[float]) -> bytes:
        """Read a sequence of bytes ending in "message_terminator".

        This method blocks until the specified message terminator sequence
        is found, then returns the received bytes including the message
        terminator.

        If "timeout" is not None and the timeout expires before the message
        terminator is received, QMI_TimeoutException is raised and any
        available bytes remain in the input buffer. If "timeout" is None,
        this method waits until the message terminator is received.

        Subclasses must override this method, if applicable.

        Parameters:
            message_terminator: Byte sequence terminating a message.
            timeout: Maximum time to wait (in seconds), or None to wait indefinitely.

        Returns:
            Received bytes, including the terminator.

        Raises:
            ~qmi.core.exceptions.QMI_TimeoutException: If the timeout expires before the
                requested number of bytes are available.
            ~qmi.core.exceptions.QMI_EndOfInputException: If the transport has been closed on
                the remote side and the end of the input is reached before the
                message terminator is found.
        """
        raise NotImplementedError("QMI_Transport.read_until not implemented")

    def read_until_timeout(self, nbytes: int, timeout: float) -> bytes:
        """Read a sequence of bytes from the transport.

        This method blocks until either the specified number of bytes
        are available or the timeout (in seconds) expires, whichever occurs
        sooner.

        If timeout occurs, the partial sequence of available bytes is returned.
        This sequence may be empty if timeout occurs before any byte was available.

        If the transport has been closed on the remote side, any remaining
        input bytes are returned (up to the maximum number of bytes requested).
        If there are no more bytes to read, QMI_EndOfInputException is raised.

        Subclasses must override this method, if applicable.

        Parameters:
            nbytes: Maximum number of bytes to read.
            timeout: Maximum time to wait (in seconds).

        Returns:
            Received bytes.

        Raises:
            ~qmi.core.exceptions.QMI_EndOfInputException: If the transport has been closed on
                the remote side and there are no more bytes to read.
        """
        raise NotImplementedError("QMI_Transport.read_until_timeout not implemented")

    def discard_read(self) -> None:
        """Discard all bytes that are immediately available for reading."""
        raise NotImplementedError("QMI_Transport.discard_read not implemented")


class TransportDescriptorParser:
    """This class is for creating a transport-specific parser classes and has (static) methods that are
    used for parsing transport strings.
    """
    def __init__(self,
                 interface: str,
                 positionals: List[Tuple[str, Tuple[Type, bool]]],
                 keywords: Mapping[str, Tuple[Type, bool]]
    ) -> None:
        self.interface = interface
        self._positionals = positionals
        self._keywords = keywords

    def parse_parameter_strings(
            self, transport_descriptor: str, default_parameters: Optional[Mapping[str, Any]] = None
    ) -> Dict[str, Any]:
        """Method for parsing transport descriptor strings.

        Parameters:
            transport_descriptor: The string to parse.
            default_parameters:   Dictionary of default parameters to be used if not present in the string.

        Returns:
            parameters:           A dictionary object of the parsed parameters.
        """
        if default_parameters is None:
            parameters = {}
        else:
            parameters = dict(default_parameters)

        # Drop unexpected default parameters.
        # These may be intended for a different transport interface.
        positional_names = {name for (name, _) in self._positionals}
        for attr in list(parameters.keys()):
            if (attr not in positional_names) and (attr not in self._keywords):
                parameters.pop(attr)

        interface = self._parse_interface(transport_descriptor)

        if self.interface != interface.lower():
            raise QMI_TransportDescriptorException(
                'Unexpected interface: {} expected {}'.format(interface, self.interface))

        parts = self._parse_parts(transport_descriptor)

        parameters.update(self._parse_positional_parameters(parts[1:]))
        parameters.update(self._parse_keyword_parameters(parts[1:]))

        self._check_missing_parameters(parameters)
        return parameters

    @staticmethod
    def _parse_parts(transport_descriptor: str) -> List[str]:
        regex = re.compile(
            r"((?:^([^:]+))|"  # transport interface: i.e. serial:...
            r"(?::\[(.+)[\]$])|"  # enclosed parameter (for example used in ipv6): i.e. ...:[param]:... or ...:[param]
            r"(?::([^:]+)))")  # regular parameter: i.e. ...:param:... or ...:param
        parts = []
        for match in re.finditer(regex, transport_descriptor):
            if match[2]:  # transport interface
                parts.append(match[2])
            elif match[3]:  # enclosed parameter
                parts.append(match[3])
            elif match[4]:  # regular parameter
                parts.append(match[4])
            else:
                raise QMI_TransportDescriptorException(
                    "Invalid transport descriptor {!r}".format(transport_descriptor))
        if len(parts) < 2:
            raise QMI_TransportDescriptorException("Invalid transport descriptor {!r}".format(transport_descriptor))
        return parts

    @staticmethod
    def _parse_interface(transport_descriptor: str) -> str:
        parts = TransportDescriptorParser._parse_parts(transport_descriptor)
        return parts[0]

    def match_interface(self, transport_descriptor: str) -> bool:
        """A method to check the transport descriptor is used with the correct parser class."""
        interface = self._parse_interface(transport_descriptor).lower()
        return self.interface == interface

    def _check_missing_parameters(self, parameters: Dict[str, Any]):
        req_params = self._get_required_parameters()
        missing_parameters = req_params.difference(parameters.keys())
        if len(missing_parameters) > 0:
            raise QMI_TransportDescriptorException('Missing required parameter(s): {}'.format(missing_parameters))

    def _parse_positional_parameters(self, params: List[str]) -> Dict[str, Any]:
        positional_params = [param for param in params if not self._is_keyword_param(param)]
        d = dict()
        for (name, (ty, _)), param in zip(self._positionals, positional_params):
            try:
                d[name] = ty(param)
            except ValueError:
                raise QMI_TransportDescriptorException("Cannot parse keyword {} expected type {} but got {}", name,
                                                       ty, param)
        return d

    def _parse_keyword_parameters(self, strings: List[str]) -> Dict[str, Any]:
        keyword_strings = [param for param in strings if self._is_keyword_param(param)]
        parameters = dict()
        for keyword_string in keyword_strings:
            q = keyword_string.split('=', maxsplit=2)
            if len(q) < 2:
                raise QMI_TransportDescriptorException('Keyword parameter is not in form of foo=bar')
            k, v = q
            if k in self._keywords.keys():
                try:
                    ty = self._keywords[k][0]
                    if ty is int and v.startswith('0x'):
                        parameters[k] = int(v, 16)
                    elif ty == bool:
                        if v not in ("True", "False"):
                            raise ValueError()
                        parameters[k] = (v == "True")
                    else:
                        parameters[k] = ty(v)
                except ValueError:
                    raise QMI_TransportDescriptorException("Cannot parse keyword {} expected type {} but got {}", k,
                                                           self._keywords[k][0], v)
            else:
                raise QMI_TransportDescriptorException("Unexpected keyword {}", k)
        return parameters

    def _get_required_parameters(self):
        required_positional = set(name for name, (_, required) in self._positionals if required)
        required_keyword = set(name for name, (_, required) in self._keywords.items() if required)
        req_params = required_positional | required_keyword
        return req_params

    @staticmethod
    def _is_keyword_param(param):
        return '=' in param


SerialTransportDescriptorParser = TransportDescriptorParser(
    "serial",
    [("device", (str, True))],
    {'baudrate': (int, False), 'bytesize': (int, False),
     'parity': (str, False),
     'stopbits': (float, False),
     'rtscts': (bool, False)}
)

TcpTransportDescriptorParser = TransportDescriptorParser(
    "tcp",
    [("host", (str, True)),
     ('port', (int, True))],
    {'connect_timeout': (float, False)}
)

UdpTransportDescriptorParser = TransportDescriptorParser(
    "udp",
    [("host", (str, True)),
     ('port', (int, True))],
    {'connect_timeout': (float, False)}
)

UsbTmcTransportDescriptorParser = TransportDescriptorParser(
    "usbtmc",
    [],
    {'vendorid': (int, False),
     'productid': (int, False),
     'serialnr': (str, True)}
)

GpibTransportDescriptorParser = TransportDescriptorParser(
    "gpib",
    [('primary_addr', (int, True))],
    {'board': (int, False),
     'secondary_addr': (int, False),
     'connect_timeout': (float, False)}
)

Vxi11TransportDescriptorParser = TransportDescriptorParser(
    "vxi11",
    [("host", (str, True))],
    {}
)


class QMI_SerialTransport(QMI_Transport):
    """Byte stream transport via serial port.

    This class can also be used for "virtual" serial ports via USB.

    Attributes:
        SERIAL_READ_TIMEOUT: Set a fixed read timeout on the serial port device. The actual specified timeout
                             for read() and read_until() calls will be rounded up to a multiple of this fixed
                             timeout. The timeout parameter of the serial port device must be fixed because
                             changing the timeout causes reprogramming of the serial port parameters,
                             which is a slow operation and can even cause data loss (with an FTDI
                             device under Windows).
    """
    SERIAL_READ_TIMEOUT = 0.040  # 40 ms

    def __init__(self,
                 device: str,
                 baudrate: int,
                 bytesize: int = 8,
                 parity: str = 'N',
                 stopbits: float = 1.0,
                 rtscts: bool = False,
                 ) -> None:
        """Create a bidirectional byte stream via a serial port.

        Parameters:
            device:   The device name, e.g. COM3 on Windows or /dev/ttyS1 or /dev/ttyUSB1 on Linux.
            baudrate: The baud rate in bits per second.
            bytesize: The number of bits per character (5, 6, 7 or 8).
            parity:   The parity mode (valid values are 'N','E','O').
            stopbits: The number of stop bits (1.0, 1.5 or 2.0).
            rtscts:   True to enable RTS/CTS flow control.
        """
        super().__init__()
        _logger.debug("Opening serial port %r baud=%d", device, baudrate)

        self._validate_device_name(device)
        self._validate_baudrate(baudrate)
        self._validate_bytesize(bytesize)
        self._validate_parity(parity)
        self._validate_stopbits(stopbits)
        self._validate_rstcts(rtscts)

        self.device = device
        self._baudrate = baudrate
        self._bytesize = bytesize
        self._parity = parity
        self._stopbits = stopbits
        self._rtscts = rtscts
        self._read_buffer = bytearray()
        self._serial: Optional[serial.Serial] = None

    @staticmethod
    def _validate_stopbits(stopbits: float) -> None:
        if stopbits not in (1.0, 1.5, 2.0):
            raise QMI_TransportDescriptorException("Invalid value for stopbits ({})".format(stopbits))

    @staticmethod
    def _validate_parity(parity: str) -> None:
        if parity not in ('N', 'E', 'O'):
            raise QMI_TransportDescriptorException("Invalid parity specification ({})".format(parity))

    @staticmethod
    def _validate_bytesize(bytesize: int) -> None:
        if bytesize < 5 or bytesize > 8:
            raise QMI_TransportDescriptorException("Invalid value for bytesize ({})".format(bytesize))

    @staticmethod
    def _validate_baudrate(baudrate: int) -> None:
        if baudrate < 1:
            raise QMI_TransportDescriptorException("Invalid baud rate ({})".format(baudrate))

    @staticmethod
    def _validate_device_name(device: str) -> None:
        if not (device.upper().startswith("COM") or device.startswith("/")):
            raise QMI_TransportDescriptorException("Unknown serial port device path ({})".format(device))

    @staticmethod
    def _validate_rstcts(rtscts: bool) -> None:
        if rtscts not in (True, False):
            raise QMI_TransportDescriptorException("Invalid rtscts ({})".format(rtscts))

    @property
    def _safe_serial(self) -> serial.Serial:
        """ The _safe_serial property should be used inside the QMI_Transport code if-and-only-if we are 100% sure that
        the _serial attribute is not None.

        This aids in static typechecking, since whereas the type of _serial is Optional[T], the result of this method
        is guaranteed to be of type T. It is a QMI-internal bug if this property is used in case _serial is None. In
        that case, we raise an AssertionError, and we hope the users will complain to us so we can fix the bug in the
        library.

        Raises: AssertionError: in case the property is used when the underlying value of _serial is None.

        Returns: The value of _serial, if it is not None. """
        assert self._serial is not None
        return self._serial

    def __str__(self) -> str:
        return "QMI_SerialTransport {!r}".format(self.device)

    def _open_transport(self) -> None:
        _logger.debug("Opening serial port %r", self.device)
        self._serial = serial.Serial(self.device,
                                     baudrate=self._baudrate,
                                     bytesize=self._bytesize,
                                     parity=self._parity,
                                     stopbits=self._stopbits,
                                     rtscts=self._rtscts,
                                     timeout=self.SERIAL_READ_TIMEOUT)

    def close(self) -> None:
        _logger.debug("Closing serial port %r", self.device)
        super().close()
        if self._serial is not None:
            self._serial.close()

    def write(self, data: bytes) -> None:
        self._check_is_open()
        self._safe_serial.write(data)

    def read(self, nbytes: int, timeout: Optional[float]) -> bytes:
        self._check_is_open()

        nbuf = len(self._read_buffer)
        if nbuf >= nbytes:
            # The requested number of bytes are already in the buffer.
            # Return them immediately.
            ret = bytes(self._read_buffer[:nbytes])
            self._read_buffer = self._read_buffer[nbytes:]
            return ret

        if (timeout is not None) and (timeout <= 0):
            # Non-blocking read requested.
            # Do not try to read from the serial port, unless we know that
            # the requested number of bytes is already available.
            if self._safe_serial.in_waiting >= nbytes - nbuf:
                self._read_buffer.extend(self._safe_serial.read(nbytes - nbuf))
        else:
            # Loop until timeout or sufficient number of bytes received.
            tstart = time.monotonic()
            while True:
                # Try to read from the serial port. This call returns when
                # the requested number of bytes are received, or when the
                # fixed serial port timeout expires.
                self._read_buffer.extend(self._safe_serial.read(nbytes - nbuf))
                nbuf = len(self._read_buffer)
                if nbuf >= nbytes:
                    break
                if timeout is not None:
                    tremain = tstart + timeout - time.monotonic()
                    if tremain <= 0:
                        break

        # Timeout before requested number of bytes received.
        # Raise exception and leave all data in the buffer.
        nbuf = len(self._read_buffer)
        if nbuf < nbytes:
            raise QMI_TimeoutException("Timeout after {} bytes while expecting {}".format(nbuf, nbytes))

        # Return the received data.
        assert nbuf == nbytes
        ret = bytes(self._read_buffer)
        self._read_buffer = bytearray()
        return ret

    def read_until(self, message_terminator: bytes, timeout: Optional[float]) -> bytes:
        self._check_is_open()

        # Check if a message terminator is already present in the buffer.
        p = self._read_buffer.find(message_terminator)
        if p < 0:
            # No message terminator in the buffer.
            # Read all bytes from the serial port that are available without waiting.
            navail = self._safe_serial.in_waiting
            self._read_buffer.extend(self._safe_serial.read(navail))

        # Check if a message terminator is already present in the buffer.
        p = self._read_buffer.find(message_terminator)
        if p >= 0:
            # Complete message already in the buffer. Return it immediately.
            nbytes = p + len(message_terminator)
            ret = bytes(self._read_buffer[:nbytes])
            self._read_buffer = self._read_buffer[nbytes:]
            return ret

        # Loop until timeout or message terminator received.
        # Do not block if a zero timeout was specified.
        tstart = time.monotonic()
        tremain = timeout
        while (tremain is None) or (tremain > 0):
            # Read a single character from the serial port.
            # This call returns when a character is received, or when
            # the fixed serial port timeout expires.
            b = self._safe_serial.read(1)
            self._read_buffer.extend(b)

            # Check if a message terminator was received.
            if self._read_buffer.endswith(message_terminator):
                # Return the complete message.
                ret = bytes(self._read_buffer)
                self._read_buffer = bytearray()
                return ret

            # Update remaining wait time.
            if timeout is not None:
                tremain = tstart + timeout - time.monotonic()

        nbuf = len(self._read_buffer)
        raise QMI_TimeoutException("Timeout after {} bytes without message terminator".format(nbuf))

    def read_until_timeout(self, nbytes: int, timeout: float) -> bytes:
        try:
            ret = self.read(nbytes, timeout)
        except QMI_TimeoutException:
            ret = bytes(self._read_buffer)
            self._read_buffer = bytearray()
        return ret

    def discard_read(self) -> None:
        self._check_is_open()
        self._safe_serial.reset_input_buffer()
        self._read_buffer = bytearray()


def _is_valid_hostname(hostname: str) -> bool:
    """Return True if the specified host name has valid syntax."""
    # source: https://stackoverflow.com/questions/2532053/validate-a-hostname-string
    if len(hostname) > 255:
        return False
    if hostname[-1] == ".":
        hostname = hostname[:-1]  # strip exactly one dot from the right, if present
    parts = hostname.split(".")
    if re.match(r"^[0-9]+$", parts[-1]):
        return False  # numeric TLD, not a domain name
    allowed = re.compile(r"(?!-)[A-Z0-9-]{1,63}(?<!-)$", re.IGNORECASE)
    return all(allowed.match(x) for x in parts)


def _is_valid_ipaddress(address: str) -> bool:
    try:
        socket.inet_pton(socket.AF_INET, address)
        return True
    except OSError:
        pass
    try:
        socket.inet_pton(socket.AF_INET6, address)
        return True
    except OSError:
        pass
    return False


class QMI_SocketTransport(QMI_Transport):
    """Base class for bidirectional data streams via socket network connection.

    Attributes:
        MIN_PACKET_SIZE: The minimum packet size to read with `read_until` method.
        MAX_PACKET_SIZE: The maximum packet size to read with `read` method.
    """
    MIN_PACKET_SIZE: int
    MAX_PACKET_SIZE: int

    def __init__(self, host: str, port: int) -> None:
        """Initialize the UDP or TCP transport with validation of host and port.

        Parameters:
            host:   The server/client IP address.
            port:   The port for the address.
        """
        super().__init__()

        # As 'localhost' does not necessarily resolve to '127.0.0.1', it is good to resolve it just-in-case.
        host = socket.gethostbyname(host) if host == "localhost" else host
        self._validate_host(host)
        self._validate_port(port)
        self._address = (host, port)
        self._socket: Optional[socket.socket] = None
        self._read_buffer = bytearray()

    @staticmethod
    def _validate_host(host):
        if (not _is_valid_hostname(host)) and (not _is_valid_ipaddress(host)):
            raise QMI_TransportDescriptorException(f"Invalid host name {host}")

    @staticmethod
    def _validate_port(port):
        if port < 1 or port > 65535:
            raise QMI_TransportDescriptorException(f"Invalid port number {port}")

    @property
    def _safe_socket(self) -> socket.socket:
        """ The _safe_socket property should be used inside the QMI_Transport code if-and-only-if we are 100% sure that
        the _socket attribute is not None.

        This aids in static typechecking, since whereas the type of _socket is Optional[T], the result of this method
        is guaranteed to be of type T. It is a QMI-internal bug if this property is used in case _socket is None. In
        that case, we raise an AssertionError, and we hope the users will complain to us so we can fix the bug in the
        library.

        Raises: AssertionError: in case the property is used when the underlying value of _socket is None.

        Returns: The value of _socket, if it is not None. """
        assert self._socket is not None
        return self._socket

    def _open_transport(self) -> None:
        _logger.debug("Opening %s with address %s", self, self._address)
        self._read_buffer = bytearray()

    def _read_from_socket(self, packet_size: int) -> Tuple[bytes, Any]:
        """Helper function to read from a socket with specific packet size.
        Exact amount does not matter but a small power of 2 is recommended, e.g. 4096
        by socket.recv() documentation.

        Parameters:
            packet_size: The packet size to read.
        """
        try:
            b, addr = self._safe_socket.recvfrom(packet_size)
        except (BlockingIOError, socket.timeout) as err:
            raise err  # Re-raise this for handling in the upper try-except.
        except OSError as err:
            raise QMI_RuntimeException(
                f"UDP packet size was larger than {packet_size}. Data is lost."
            ) from err

        if not b:
            raise QMI_EndOfInputException(
                f"Reached end of input from socket {format_address_and_port(self._address)}"
            )
        return b, addr

    def close(self) -> None:
        super().close()
        self._safe_socket.close()

    def write(self, data: bytes) -> None:
        raise NotImplementedError("QMI_SocketTransportBase.write not implemented")

    def read(self, nbytes: int, timeout: Optional[float] = None) -> bytes:
        self._check_is_open()
        tstart = time.monotonic()
        tremain = timeout
        nbuf = len(self._read_buffer)
        while nbuf < nbytes:
            self._safe_socket.settimeout(tremain)
            try:
                b, addr = self._read_from_socket(max(nbytes - nbuf, self.MIN_PACKET_SIZE))

            except (BlockingIOError, socket.timeout) as err:
                raise QMI_TimeoutException(
                    f"Timeout after {nbuf} bytes while expecting {nbytes}"
                ) from err
            self._read_buffer.extend(b)
            nbuf = len(self._read_buffer)
            if timeout is not None:
                tremain = tstart + timeout - time.monotonic()
                if tremain < 0:
                    raise QMI_TimeoutException(f"Timeout after {nbuf} bytes while expecting {nbytes}")

        ret = bytes(self._read_buffer[:nbytes])
        self._read_buffer = self._read_buffer[nbytes:]
        return ret

    def read_until(self, message_terminator: bytes, timeout: Optional[float] = None) -> bytes:
        # Check if message terminator already received. To be further handled in subclass implementation.
        p = self._read_buffer.find(message_terminator)
        if p >= 0:
            # Found "message_terminator" - return data.
            nbytes = p + len(message_terminator)
            ret = bytes(self._read_buffer[:nbytes])
            self._read_buffer = self._read_buffer[nbytes:]
            return ret

        self._check_is_open()
        tstart = time.monotonic()
        tremain = timeout
        while True:
            self._safe_socket.settimeout(tremain)
            try:
                # Read from socket.
                b, addr = self._read_from_socket(self.MAX_PACKET_SIZE)
                self._read_buffer.extend(b)
            except (BlockingIOError, socket.timeout, TimeoutError) as err:
                nbuf = len(self._read_buffer)
                raise QMI_TimeoutException(
                    f"Socket timeout after {nbuf} bytes without message terminator"
                ) from err

            # Check if message terminator received.
            p = self._read_buffer.find(message_terminator)
            if p >= 0:
                # Found "message_terminator" - return data.
                nbytes = p + len(message_terminator)
                ret = bytes(self._read_buffer[:nbytes])
                self._read_buffer = self._read_buffer[nbytes:]
                return ret

            # Determine remaining time if message terminator is not yet received.
            if timeout is not None:
                # Calculate remaining time.
                tremain = tstart + timeout - time.monotonic()
                if tremain < 0:
                    nbuf = len(self._read_buffer)
                    raise QMI_TimeoutException(f"Function timeout after {nbuf} bytes without message terminator.")

    def read_until_timeout(self, nbytes: int, timeout: float) -> bytes:
        try:
            ret = self.read(nbytes, timeout)
        except QMI_TimeoutException:
            ret = bytes(self._read_buffer)
            self._read_buffer = bytearray()
        except QMI_EndOfInputException:
            if not self._read_buffer:
                raise
            ret = bytes(self._read_buffer)
            self._read_buffer = bytearray()
        return ret

    def discard_read(self) -> None:
        self._check_is_open()
        self._read_buffer = bytearray()
        self._safe_socket.settimeout(0)
        while True:
            try:
                b = self._safe_socket.recv(self.MAX_PACKET_SIZE)
            except (BlockingIOError, socket.timeout):
                # no more bytes available
                break
            except OSError:
                # UDP protocol was used and > 4096 bytes in socket buffer. This discards also the rest of the packet.
                break

            if not b:
                # end of stream
                break


class QMI_UdpTransport(QMI_SocketTransport):
    """
    An instance of QMI_UdpTransport represents a server-side UDP connection with
    a listening port to an instrument. Client-side UDP connections are not supported.

    The maximum UDP packet size is 64KB (including headers). For `write` and `read` functions,
    this means limitations. The write function we can check the data size to send and split into
    multiple messages. For `read` functions this is not possible and the client must limit the
    data size as necessary. If for `read` functions the receivable data size is larger than
    the size defined in `recvfrom(<size>)`, an exception is raised.

    Attributes:
        MIN_PACKET_SIZE: The minimum packet size to read with `read` method.
        MAX_PACKET_SIZE: The maximum packet size to read with `read_until` method.
    """
    MIN_PACKET_SIZE = 4096
    MAX_PACKET_SIZE = 4096

    def __init__(self, host: str, port: int) -> None:
        """Initialize the UDP transport by validating the UDP host and port.

        Parameters:
            host: The server IP address.
            port: The port for the address.
        """
        super().__init__(host, port)

    def _validate_port(self, port):
        super()._validate_port(port)
        if port == QMI_Context.DEFAULT_UDP_RESPONDER_PORT:
            raise QMI_TransportDescriptorException(f"UDP port number {port} not allowed")

    def __str__(self) -> str:
        remote_addr = format_address_and_port(self._address)
        return f"QMI_UdpTransport(remote={remote_addr})"

    def _open_transport(self) -> None:
        super()._open_transport()
        # Validate the host exists before trying to create a socket
        socket.gethostbyname(self._address[0])
        # Create socket
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        # To set the local address to point to our client, we need to bind it.
        self._socket.bind(("", self._address[1]))

    def close(self) -> None:
        _logger.debug("Closing UDP transport %s", self)
        super().close()

    def write(self, data: bytes) -> None:
        self._check_is_open()
        # NOTE: We explicitly adjust the socket timeout before each send/recv call.
        self._safe_socket.settimeout(None)
        self._safe_socket.sendto(data, self._address)


class QMI_TcpTransport(QMI_SocketTransport):
    """Bidirectional byte stream via TCP network connection.

    An instance of QMI_TcpTransport represents a client-side TCP connection
    to an instrument. Server-side TCP connections are not supported.

    Attributes:
        DEFAULT_CONNECT_TIMEOUT: A default timeout period for connecting to TCP client. Default is 10 seconds.
        MIN_PACKET_SIZE: The minimum packet size to read with `read` method.
        MAX_PACKET_SIZE: The maximum packet size to read with `read_until` method.
    """

    DEFAULT_CONNECT_TIMEOUT = 10
    MIN_PACKET_SIZE = 0
    MAX_PACKET_SIZE = 512

    def __init__(self, host: str, port: int, connect_timeout: Optional[float] = DEFAULT_CONNECT_TIMEOUT) -> None:
        """Initialize the TCP transport by connecting to the specified address.

        Parameters:
            connect_timeout: Maximum time to connect in seconds.

        Raises:
            ~qmi.core.exceptions.QMI_TimeoutException: If connecting takes longer than the specified connection
            timeout.
        """
        super().__init__(host, port)
        self._connect_timeout = connect_timeout

    def __str__(self) -> str:
        remote_addr = format_address_and_port(self._address)
        return f"QMI_TcpTransport(remote={remote_addr})"

    def _open_transport(self) -> None:
        super()._open_transport()
        # Create socket and connect.
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.settimeout(self._connect_timeout)
        # Set TCP_NODELAY socket option.
        self._socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        try:
            self._socket.connect(self._address)
        except socket.timeout as e:
            self._socket.close()
            raise QMI_TimeoutException("Timeout while connecting to {}".format(self._address)) from e

    def close(self) -> None:
        _logger.debug("Closing TCP transport %s", self)
        super().close()

    def write(self, data: bytes) -> None:
        self._check_is_open()
        # NOTE: We explicitly adjust the socket timeout before each send/recv call.
        self._safe_socket.settimeout(None)
        self._safe_socket.sendall(data)


class QMI_UsbTmcTransport(QMI_Transport):
    """Transport SCPI commands via USBTMC device class.

    When running under Windows, this class depends on VISA for the actual
    USBTMC implementation.

    When running under Linux, this class depends on python-usbtmc (and libusb)
    for the actual USBTMC implementation.

    This class does not implement the full functionality of QMI_Transport.
    The issue is that USBTMC is fundamentally a message-oriented protocol,
    while QMI_Transport assumes a byte stream without message delimiters.

    Only the following operations are supported:
      * write() writes the specified bytes as a single USBTMC message.
      * read_until() reads a single USBTMC message (until the device indicates
        end-of-message) and returns the fetched bytes.

    Attributes:
        DEFAULT_READ_TIMEOUT: Default timeout in seconds for USBTMC read transactions.
        WRITE_TIMEOUT:        Timeout in seconds for USBTMC write transactions.
    """

    DEFAULT_READ_TIMEOUT = 60
    WRITE_TIMEOUT = 5

    def __init__(self, vendorid: int, productid: int, serialnr: str) -> None:
        """Initialize te specified USB device as USBTMC instrument.

        The first USBTMC-compatible interface of the USB device will be used.

        Parameters:
            vendorid: USB vendor ID.
            productid: USB product ID.
            serialnr: USB device serial number.
        """
        _logger.debug("Opening USBTMC device 0x%04x:0x%04x (%s)", vendorid, productid, serialnr)
        super().__init__()

        self._validate_vendor_id(vendorid)
        self._validate_product_id(productid)

        self.vendorid = vendorid
        self.productid = productid
        self.serialnr = serialnr

    @staticmethod
    def _validate_product_id(productid):
        if (productid < 0) or (productid > 65535):
            raise QMI_TransportDescriptorException("Missing/bad USB product ID {}".format(productid))

    @staticmethod
    def _validate_vendor_id(vendorid):
        if (vendorid < 0) or (vendorid > 65535):
            raise QMI_TransportDescriptorException("Missing/bad USB vendor ID {}".format(vendorid))

    def __str__(self) -> str:
        return f"QMI_UsbTmcTransport 0x{self.vendorid:04x}:0x{self.productid:04x} ({self.serialnr})"

    def close(self) -> None:
        _logger.debug("Closing USBTMC device 0x%04x:0x%04x (%s)", self.vendorid, self.productid, self.serialnr)
        super().close()

    def write(self, data: bytes) -> None:
        """Send the specified data to the instrument as a single USBTMC message. A fixed timeout of 5 seconds
        is applied. If timeout occurs, the transfer is aborted and an exception is raised.

        Parameters:
            data: The data to write in a byte string.
        """
        raise NotImplementedError("QMI_UsbTmcTransport.write not implemented")

    def read(self, nbytes: int, timeout: Optional[float]) -> bytes:
        """Read a specified number of bytes from the transport.

        All bytes must belong to the same USBTMC message.

        This method blocks until the specified number of bytes are available,
        then returns the received bytes. If timeout occurs, any partial read
        data is discarded and QMI_TimeoutException is raised.

        Parameters:
            nbytes: Expected number of bytes to read.
            timeout: Maximum time to wait in seconds (default: 60 seconds).

        Returns:
            Received bytes.

        Raises:
            ~qmi.core.exceptions.QMI_TimeoutException: If the timeout expires before the
                requested number of bytes are available.
            ~qmi.core.exceptions.QMI_EndOfInputException: If an end-of-message indicator
                is received before the requested number of bytes is reached.
        """
        self._check_is_open()
        # USB requires a timeout
        if timeout is None:
            timeout = self.DEFAULT_READ_TIMEOUT

        # Read a USB message
        ret = self._read_message(timeout)
        if len(ret) == nbytes:
            return ret

        _logger.debug("USBTMC read message contained data %s" % ret.decode())
        raise QMI_EndOfInputException(
            f"The read message did not contain expected bytes of data ({len(ret)} != {nbytes}."
        )

    def read_until(self, message_terminator: bytes, timeout: Optional[float]) -> bytes:
        """The specified message_terminator is ignored. Instead, the instrument must autonomously indicate the end
        of the message according to the USBTMC protocol.

        As the message_terminator is not used, we forward simply the call to the `read_until_timeout` call.

        Parameters:
            message_terminator: This input is ignored.
            timeout: Maximum time to wait (in seconds).

        Returns:
            Received bytes.
        """
        return self.read_until_timeout(0, timeout)

    def read_until_timeout(self, nbytes: int, timeout: Optional[float]) -> bytes:
        """Read a single USBTMC message from the instrument.

        If the timeout expires before the message is received, the read is
        aborted and any data already received are discarded. In this
        case an empty bytes string is returned.

        Parameters:
            nbytes: This input is ignored.
            timeout: Maximum time to wait (in seconds).

        Returns:
            Received bytes.
        """
        self._check_is_open()

        # USB requires a timeout
        if timeout is None:
            timeout = self.DEFAULT_READ_TIMEOUT

        # Read a new message from the instrument.
        try:
            data = self._read_message(timeout)
        except QMI_TimeoutException:
            data = bytes()

        return data

    def discard_read(self) -> None:
        try:
            self._read_message(0.0)
        except QMI_TimeoutException:
            return  # Nothing was in the instrument buffer, so we just continue.

    def _read_message(self, timeout):
        """Read one USBTMC message from the instrument.
        This should be implemented in deriving classes.

        Parameters:
            timeout: Obligatory parameter to set the timeout for reading.

        Raises:
            QMI_TimeoutException: If the read is not done within timeout period.
        """
        raise NotImplementedError("QMI_UsbTmcTransport.read not implemented")

    @staticmethod
    def _format_resources(resources: List[str]) -> List[str]:
        transports = set()
        for res in resources:
            parts = res.split("::")
            if len(parts) >= 5 and parts[0].startswith("USB") and parts[-1] == "INSTR":
                try:
                    vendorid = int(parts[1], 0)
                    productid = int(parts[2], 0)
                    serialnr = parts[3]
                    transports.add("usbtmc:vendorid=0x{:04x}:productid=0x{:04x}:serialnr={}"
                                   .format(vendorid, productid, serialnr))
                except ValueError:
                    pass  # ignore malformed resource strings

        return sorted(list(transports))

    @staticmethod
    def list_resources() -> List[str]:
        """List available resources from USB connections.
        This should be implemented in deriving classes.

        Returns:
            resources: List of available resources in QMI's USBTMC transport string format.
        """
        raise NotImplementedError("QMI_UsbTmcTransport.list_resources not implemented")


class QMI_Vxi11Transport(QMI_Transport):
    """ VXI-11 based transport to a supported device. """

    def __init__(self, host: str) -> None:
        """ Initialize the VXI-11 transport by connecting to the specified host.

        Parameters:
            host: Host name or address of the device.

        Raises:
            ~qmi.core.exceptions.QMI_TransportDescriptorException: If the host string is invalid.
        """
        super().__init__()

        QMI_TcpTransport._validate_host(host)

        self._host = host
        self._instr: Optional[vxi11.Instrument] = None
        self._read_buffer = bytes()

    @property
    def _safe_instr(self) -> vxi11.Instrument:
        """ The _safe_instr property should be used inside the QMI_Transport code if-and-only-if we are 100% sure that
        the _instr attribute is not None.

        This aids in static typechecking, since whereas the type of _instr is Optional[T], the result of this method is
        guaranteed to be of type T. It is a QMI-internal bug if this property is used in case _instr is None. In that
        case, we raise an AssertionError, and we hope the users will complain to us, so we can fix the bug in the
        library.

        Raises: AssertionError: in case the property is used when the underlying value of _instr is None.

        Returns: The value of _instr, if it is not None. """
        assert self._instr is not None
        return self._instr

    def _open_transport(self) -> None:
        _logger.debug("Opening %s", self)

        # Open actual transport.
        try:
            self._instr = vxi11.Instrument(self._host)
            self._safe_instr.open()
        except vxi11.vxi11.Vxi11Exception as err:
            raise QMI_InstrumentException("Error attempting to open VXI11 transport to {}".format(self._host)) from err

    def close(self) -> None:
        _logger.debug("Closing %s", self)
        super().close()
        try:
            self._safe_instr.close()
        except vxi11.vxi11.Vxi11Exception as err:
            raise QMI_InstrumentException("Error attempting to close VXI11 transport.") from err

    def write(self, data: bytes) -> None:
        self._check_is_open()
        try:
            self._safe_instr.write_raw(data)
        except vxi11.vxi11.Vxi11Exception as err:
            raise QMI_InstrumentException() from err

    def read(self, nbytes: int, timeout: Optional[float] = None) -> bytes:
        self._check_is_open()
        nbuf = len(self._read_buffer)
        if nbuf >= nbytes:
            # The requested number of bytes are already in the buffer. Return them immediately.
            ret = self._read_buffer[:nbytes]
            self._read_buffer = self._read_buffer[nbytes:]
            return ret

        old_timeout = self._safe_instr.timeout
        if timeout:
            self._safe_instr.timeout = timeout

        try:
            while len(self._read_buffer) < nbytes:
                self._read_buffer += self._safe_instr.read_raw(self._safe_instr.max_recv_size)

        except vxi11.vxi11.Vxi11Exception as err:
            if err.err == 15:
                # Raise timeout exception separately to provide opportunity for upper layer to handle it.
                raise QMI_TimeoutException(f"Timeout error attempting to read {nbytes} bytes") from err
            else:
                raise QMI_EndOfInputException(f"Error attempting to read {nbytes} bytes.") from err

        finally:
            self._safe_instr.timeout = old_timeout

        ret = self._read_buffer
        self._read_buffer = bytes()
        return ret

    def read_until(self, message_terminator: bytes, timeout: Optional[float] = None) -> bytes:
        self._check_is_open()
        if len(message_terminator) != 1:
            raise QMI_InstrumentException(
                "VXI11 instrument only support 1 byte terminating character, received {!r}.".format(message_terminator)
            )

        nbuf = len(self._read_buffer)
        if nbuf > 0 and message_terminator in self._read_buffer:
            # The requested response is already in the buffer. Return it immediately.
            terminator_index = self._read_buffer.index(message_terminator)
            ret = self._read_buffer[:terminator_index]
            self._read_buffer = self._read_buffer[terminator_index:]
            return ret

        # Set terminator, but keep old value.
        old_term_char = self._safe_instr.term_char
        self._safe_instr.term_char = message_terminator

        # Set timeout if available.
        old_timeout = self._safe_instr.timeout
        if timeout:
            self._safe_instr.timeout = timeout

        try:
            while True:
                self._read_buffer += self._safe_instr.read_raw(self._safe_instr.max_recv_size)
                # Validate terminator.
                data_term_char = self._read_buffer[-1:]  # use slice rather than index to get back bytes()
                if data_term_char == message_terminator:
                    break

        except vxi11.vxi11.Vxi11Exception as err:
            if err.err == 15:
                # Raise timeout exception separately to provide opportunity for upper layer to handle it.
                raise QMI_TimeoutException("Timeout error attempting to read bytes until {!r} char received".format(
                    message_terminator)) from err
            else:
                raise QMI_InstrumentException("Error attempting to read bytes until {!r} char received".format(
                    message_terminator)) from err

        finally:
            # Revert terminator and timeout to previous value
            self._safe_instr.term_char = old_term_char
            self._safe_instr.timeout = old_timeout

        ret = self._read_buffer
        self._read_buffer = bytes()
        return ret

    def read_until_timeout(self, nbytes: int, timeout: float) -> bytes:
        try:
            return self.read(nbytes, timeout)

        except QMI_TimeoutException:
            # Return whatever was read until timeout and clear the buffer.
            ret = self._read_buffer
            self._read_buffer = bytes()
            return ret

    def discard_read(self) -> None:
        self._check_is_open()
        old_timeout = self._safe_instr.timeout
        self._safe_instr.timeout = 0.0  # Immediate read
        # Clear any possible data in read buffer first
        self._read_buffer = bytes()
        while True:
            try:
                self._safe_instr.read_raw(1)

            except vxi11.vxi11.Vxi11Exception as err:
                if err.err == 15:
                    break
                else:
                    self._safe_instr.timeout = old_timeout
                    raise QMI_InstrumentException("Error attempting to read a byte") from err

            except TimeoutError:
                break

        self._safe_instr.timeout = old_timeout


def list_usbtmc_transports() -> List[str]:
    """Return a list of currently detected USBTMC transports."""
    if sys.platform.lower().startswith("win"):
        from qmi.core.transport_usbtmc_visa import QMI_VisaUsbTmcTransport
        return QMI_VisaUsbTmcTransport.list_resources()
    else:
        # On Linux, we use a copy of python-usbtmc which is integrated in QMI.
        from qmi.core.transport_usbtmc_pyusb import QMI_PyUsbTmcTransport
        return QMI_PyUsbTmcTransport.list_resources()


def create_transport(
        transport_descriptor: str, default_attributes: Optional[Dict[str, Any]] = None
) -> QMI_Transport:
    """Create a bidirectional communication channel.

    A transport_descriptor specifies all information that may be needed to open a transport, including parameters
    such as port number, baud rate, etc. Certain entries are obligatory, like giving the host IP address for UDP and
    TCP transports. Other entries are optional, and are indicated with `<`, `>` characters. For those entries, if
    not given, the string format below indicates the default value used in that case with the `=value` part. Do not
    include the `<`, `>` characters in the strings.

    String format:
      - VXI-11 instrument: "vxi11:host"
      - UDP connection:    "udp:host<:port>"
      - TCP connection:    "tcp:host<:port><:connect_timeout=10>"
      - Serial port:       "serial:device<:baudrate=115200><:databits=8><:parity=N><:stopbits=1>"
      - USBTMC device:     "usbtmc:vendorid:productid:serialnr"
      - GPIB device:       "gpib:<board=None:>primary_addr<:secondary_addr=None><:connect_timeout=30.0>"

    UDP, TCP and VXI-11:
      - "host" (for UDP, TCP & VXI-11 transports) specifies the host name or IP address of the UDP server/TCP client.
        Numerical IPv6 addresses must be enclosed in square brackets, e.g. "tcp:[2620:0:2d0:200::8]:5000".
      - "port" (for UDP and TCP transports) specifies the UDP/TCP port number of the server/client.
      - "connect_timeout" is TCP connection timeout.

     Serial:
      - "device" is the name of the serial port, for example "COM3" or "/dev/ttyUSB0".
      - "baudrate" specifies the number of bits per second.
        This attribute is only required for instruments with a configurable baud rate.
      - "bytesize" specifies the number of data bits per character (valid range 5 - 8).
        This attribute is only required for instruments with a configurable character format.
      - "parity" specifies the parity bit ('O' or 'E' or ''N').
        This attribute is only required for instruments with a configurable character format.
      - "stopbits" specifies the number of stop bits (1 or 1.5 or 2).
        This attribute is only required for instruments with a configurable character format.
      - "rtscts" enables or disables RTS/CTS flow control.
        Possible values are True and False; the default is False.

    USBTMC:
      - "vendorid" is the USB Vendor ID as a decimal number or as hexadecimal with 0x prefix.
      - "productid" is the USB Product ID as a decimal number or as hexadecimal with 0x prefix.
      - "serialnr" is the USB serial number string.

    GPIB:
      - "primary_addr" is GPIB device number (integer).
      - "board" is optional GPIB interface number (in VISA syntax GPIB[board]::...).
      - "secondary_addr" is optional secondary device address number.
      - "connect_timeout" is for opening resource for GPIB device, in seconds.
    """
    if SerialTransportDescriptorParser.match_interface(transport_descriptor):
        attributes = SerialTransportDescriptorParser.parse_parameter_strings(transport_descriptor, default_attributes)
        return QMI_SerialTransport(**attributes)
    elif UdpTransportDescriptorParser.match_interface(transport_descriptor):
        attributes = UdpTransportDescriptorParser.parse_parameter_strings(transport_descriptor, default_attributes)
        return QMI_UdpTransport(**attributes)
    elif TcpTransportDescriptorParser.match_interface(transport_descriptor):
        attributes = TcpTransportDescriptorParser.parse_parameter_strings(transport_descriptor, default_attributes)
        return QMI_TcpTransport(**attributes)
    elif UsbTmcTransportDescriptorParser.match_interface(transport_descriptor):
        attributes = UsbTmcTransportDescriptorParser.parse_parameter_strings(transport_descriptor, default_attributes)
        if sys.platform.lower().startswith("win"):
            from qmi.core.transport_usbtmc_visa import QMI_VisaUsbTmcTransport
            return QMI_VisaUsbTmcTransport(**attributes)
        else:
            # On Linux, we use a copy of python-usbtmc which is integrated in QMI.
            from qmi.core.transport_usbtmc_pyusb import QMI_PyUsbTmcTransport
            return QMI_PyUsbTmcTransport(**attributes)

    elif GpibTransportDescriptorParser.match_interface(transport_descriptor):
        attributes = GpibTransportDescriptorParser.parse_parameter_strings(transport_descriptor, default_attributes)
        if sys.platform.lower().startswith("win"):
            from qmi.core.transport_gpib_visa import QMI_VisaGpibTransport
            return QMI_VisaGpibTransport(**attributes)
        else:
            # This is a Windows-specific transport for National Instruments GPIB-USB-HS.
            raise QMI_TransportDescriptorException(
                "Gpib transport descriptor is for NI GPIB-USB-HS device and Windows-only."
            )

    elif Vxi11TransportDescriptorParser.match_interface(transport_descriptor):
        attributes = Vxi11TransportDescriptorParser.parse_parameter_strings(transport_descriptor, default_attributes)
        return QMI_Vxi11Transport(**attributes)
    else:
        raise QMI_TransportDescriptorException("Unknown type in transport descriptor %r" % transport_descriptor)



"""
Montana Instruments, cryostations.

The qmi.instruments.montana package provides support for:
- Cryostation S50 with RESTful API
- Cryostations through Montana laptop
"""

from qmi.instruments.montana.cryostation import Montana_Cryostation
from qmi.instruments.montana.cryostation_s50 import Montana_CryostationS50, \
    Montana_CryostationS50_Thermometer_Properties



#! /usr/bin/env python3

import logging
import time
import unittest
from unittest.mock import ANY

from qmi.core.config_defs import CfgQmi, CfgContext
from qmi.core.context import QMI_Context
from qmi.core.messaging import QMI_Message, QMI_MessageHandler, QMI_MessageHandlerAddress
from qmi.core.exceptions import QMI_MessageDeliveryException


logging.getLogger().setLevel(logging.CRITICAL)


class TestMessage(QMI_Message):
    """Basic test message."""
    def __init__(self, source_address: QMI_MessageHandlerAddress, destination_address: QMI_MessageHandlerAddress, value):
        super().__init__(source_address, destination_address)
        self.value = value


class TestMessageHandler(QMI_MessageHandler):
    """Message handler mock for testing purposes."""
    def __init__(self, name):
        super().__init__(name)
        self.received = set()
        self.last_message = None

    def handle_message(self, message):
        assert message.value not in self.received
        self.received.add(message.value)
        self.last_message = message


class TestBasicMessaging(unittest.TestCase):
    """Test basic message sending and handling between contexts."""

    def setUp(self):
        # Start three contexts.

        config = CfgQmi(
            contexts={
                "c1": CfgContext(tcp_server_port=0),
                "c2": CfgContext(tcp_server_port=0),
                "c3": CfgContext(tcp_server_port=0)
            }
        )

        c1 = QMI_Context("c1", config)
        c1.start()
        c1_port = c1.get_tcp_server_port()

        c2 = QMI_Context("c2", config)
        c2.start()
        c2_port = c2.get_tcp_server_port()

        c3 = QMI_Context("c3", config)
        c3.start()
        c3_port = c3.get_tcp_server_port()

        # Note: Connecting to "localhost" on Windows can give 1 second delay
        # because IPv6 is attempted before IPv4.
        # This problem does not occur when connecting to "127.0.0.1".

        # Connect c1 to c2 and c3.
        c1.connect_to_peer("c2", "localhost:{}".format(c2_port))
        c1.connect_to_peer("c3", "127.0.0.1:{}".format(c3_port))

        # Connect c2 to c1.
        c2.connect_to_peer("c1", "127.0.0.1:{}".format(c1_port))

        self.c1 = c1
        self.c2 = c2
        self.c3 = c3

    def tearDown(self):
        self.c1.stop()
        self.c2.stop()
        self.c3.stop()

        self.c1 = None
        self.c2 = None
        self.c3 = None

    def test_delivery(self):
        """Basic message delivery."""
        mh1 = TestMessageHandler(QMI_MessageHandlerAddress("c1", "mh1"))
        mh2 = TestMessageHandler(QMI_MessageHandlerAddress("c2", "mh2"))
        mh3 = TestMessageHandler(QMI_MessageHandlerAddress("c3", "mh3"))

        self.c1.register_message_handler(mh1)
        self.c2.register_message_handler(mh2)
        self.c3.register_message_handler(mh3)

        expect_1 = set()
        expect_2 = set()
        expect_3 = set()

        for r in range(1, 4):
            if r == 1:
                sendctx = self.c2
                sender = mh2
                receiver = mh1
                expect_1.add(r)
            elif r == 2:
                sendctx = self.c1
                sender = mh1
                receiver = mh2
                expect_2.add(r)
            elif r == 3:
                sendctx = self.c1
                sender = mh1
                receiver = mh3
                expect_3.add(r)

            msg = TestMessage(sender.address, receiver.address, r)
            sendctx.send_message(msg)

        time.sleep(0.3)

        self.c1.unregister_message_handler(mh1)
        self.c2.unregister_message_handler(mh2)
        self.c3.unregister_message_handler(mh3)

        self.assertEqual(expect_1, mh1.received)
        self.assertEqual(expect_2, mh2.received)
        self.assertEqual(expect_3, mh3.received)

    def test_unroutable_message(self):
        """Undeliverable messages."""
        mh1 = TestMessageHandler(QMI_MessageHandlerAddress("c1", "mh1"))
        mh3 = TestMessageHandler(QMI_MessageHandlerAddress("c3", "mh3"))

        self.c1.register_message_handler(mh1)
        self.c3.register_message_handler(mh3)

        # c3 can not send messages directly to c1 (because it never connected to c1)
        msg = TestMessage(mh3.address, mh1.address, 1)
        with self.assertRaises(QMI_MessageDeliveryException):
            self.c3.send_message(msg)

        self.c1.unregister_message_handler(mh1)
        self.c3.unregister_message_handler(mh3)

    def test_local_alias(self):
        """Local aliases."""
        mh1 = TestMessageHandler(QMI_MessageHandlerAddress("c1", "mh1"))
        mh3 = TestMessageHandler(QMI_MessageHandlerAddress("c3", "mh3"))

        self.c1.register_message_handler(mh1)
        self.c3.register_message_handler(mh3)

        # c1 can send directly to c3
        msg = TestMessage(mh1.address, mh3.address, 1)
        self.c1.send_message(msg)

        time.sleep(0.1)

        # c3 will receive the message with source address renamed to its local alias for c1
        self.assertIsNotNone(mh3.last_message)
        self.assertTrue(mh3.last_message.source_address.context_id.startswith("$"))
        self.assertEqual(mh3.last_message.source_address.object_id, mh1.address.object_id)

        # c3 can send a message to its local alias for c1
        msg = TestMessage(mh3.address, mh3.last_message.source_address, 2)
        self.c3.send_message(msg)

        time.sleep(0.1)

        # c1 will receive the message
        self.assertIsNotNone(mh1.last_message)
        self.assertEqual(mh1.last_message.source_address, mh3.address)

        self.c1.unregister_message_handler(mh1)
        self.c3.unregister_message_handler(mh3)


class TestUdpMessaging(unittest.TestCase):
    """Test UDP message handling."""

    def setUp(self):
        # Create two contexts.
        configA = CfgQmi(
            workgroup="qmi_ut_wgA",
            contexts={
                "foo": CfgContext(tcp_server_port=60001),
                "bar": CfgContext(tcp_server_port=60002)
            }
        )

        configB = CfgQmi(
            workgroup="qmi_ut_wgB",
            contexts={
                "baz": CfgContext(tcp_server_port=60003)
            }
        )

        ctx_foo = QMI_Context("foo", configA)
        ctx_foo.start()

        ctx_bar = QMI_Context("ctx_bar", configA)
        ctx_bar.start()

        ctx_baz = QMI_Context("ctx_baz", configB)
        ctx_baz.start()

        self.ctx_foo = ctx_foo
        self.ctx_bar = ctx_bar
        self.ctx_baz = ctx_baz

    def tearDown(self):
        self.ctx_foo.stop()
        self.ctx_bar.stop()
        self.ctx_baz.stop()

        self.ctx_foo = None
        self.ctx_bar = None
        self.ctx_baz = None

    def test_discovery_default(self):
        """Basic discovery with default options. It should find only the contexts with the same workgroup name as
        ctx_foo itself has (the default workgroup name is its own context's workgroup name)."""
        peers = self.ctx_foo.discover_peer_contexts()
        self.assertNotIn(("ctx_foo", ANY), peers)
        self.assertIn(("ctx_bar", ANY), peers)
        self.assertNotIn(("ctx_baz", ANY), peers)

    def test_discovery_all(self):
        """Discovery with setting key character '*' to find all workgroup names"""
        peers = self.ctx_foo.discover_peer_contexts(workgroup_name_filter='*')
        self.assertNotIn(("ctx_foo", ANY), peers)  # That's me, not a peer
        self.assertIn(("ctx_bar", ANY), peers)
        self.assertIn(("ctx_baz", ANY), peers)

    def test_discovery_workgroup_filter1(self):
        """Test workgroup filtering."""
        peers = self.ctx_foo.discover_peer_contexts(workgroup_name_filter="qmi_ut_wg?")
        self.assertNotIn(("ctx_foo", ANY), peers)
        self.assertIn(("ctx_bar", ANY), peers)
        self.assertIn(("ctx_baz", ANY), peers)

    def test_discovery_workgroup_filter2(self):
        """Test workgroup filtering."""
        peers = self.ctx_foo.discover_peer_contexts(workgroup_name_filter="qmi_ut_*")
        self.assertNotIn(("ctx_foo", ANY), peers)
        self.assertIn(("ctx_bar", ANY), peers)
        self.assertIn(("ctx_baz", ANY), peers)

    def test_discovery_workgroup_filter3(self):
        """Test workgroup filtering."""
        peers = self.ctx_foo.discover_peer_contexts(workgroup_name_filter="qmi_ut_*A")
        self.assertNotIn(("ctx_foo", ANY), peers)
        self.assertIn(("ctx_bar", ANY), peers)
        self.assertNotIn(("ctx_baz", ANY), peers)

    def test_discovery_context_filter1(self):
        """Test context filtering."""
        peers = self.ctx_foo.discover_peer_contexts(workgroup_name_filter='*', context_name_filter="ctx_ba?")
        self.assertNotIn(("ctx_foo", ANY), peers)
        self.assertIn(("ctx_bar", ANY), peers)
        self.assertIn(("ctx_baz", ANY), peers)

    def test_discovery_context_filter2(self):
        """Test context filtering."""
        peers = self.ctx_foo.discover_peer_contexts(workgroup_name_filter='*', context_name_filter="ctx_b*")
        self.assertNotIn(("ctx_foo", ANY), peers)
        self.assertIn(("ctx_bar", ANY), peers)
        self.assertIn(("ctx_baz", ANY), peers)

    def test_discovery_context_filter3(self):
        """Test context filtering."""
        peers = self.ctx_baz.discover_peer_contexts(workgroup_name_filter='*', context_name_filter="ctx_b*")
        self.assertNotIn(("ctx_foo", ANY), peers)
        self.assertIn(("ctx_bar", ANY), peers)
        self.assertNotIn(("ctx_baz", ANY), peers)

    def test_discovery_context_filter4(self):
        """Test context filtering."""
        peers = self.ctx_foo.discover_peer_contexts(workgroup_name_filter='*', context_name_filter="ctx_bar")
        self.assertNotIn(("ctx_foo", ANY), peers)
        self.assertIn(("ctx_bar", ANY), peers)
        self.assertNotIn(("ctx_baz", ANY), peers)


if __name__ == "__main__":
    unittest.main()



"""QMI driver for the Agiltron FF 1x4 optical switch driver.

The protocol implemented in this driver is documented in the document
'Command_List_for_CL-LB_Switch_Driver_5-20-2020.doc'

To use the instrument on Windows, it is necessary to have Virtual COM Port (VCP) driver installed on the PC.
See https://ftdichip.com/Drivers/vcp-drivers/.

"""

import logging
from typing import List, Optional

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import rpc_method
from qmi.core.transport import create_transport

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class Agiltron_FF1x4(QMI_Instrument):
    """QMI driver for the Agiltron FF 1x4 optical switch driver.
    driver-specific constants:
    DEFAULT_TIMEOUT: float. The default time-out for the optical switch serial commands
    CHANNELS: int. The number of channels in the switch.
    """
    DEFAULT_TIMEOUT = 0.500  # 500 ms ought to be plenty.
    CHANNELS = 4

    def __init__(self, context: QMI_Context, name: str, transport: str, timeout: Optional[float] = None):
        """

        :param context: QMI_Context object for the instrument driver.
        :param name: string. Name for this instrument instance.
        :param transport: string. Transport descriptor to access the instrument.
        :param timeout: Optional[float]. Alternative timeout for the default instrument serial communication.
        """
        super().__init__(context, name)
        self._transport = create_transport(
            transport,
            default_attributes={"baudrate": 9600, "bytesize": 8, "parity": "N", "stopbits": 1, "rtscts": False},
        )

        if timeout is None:
            timeout = self.DEFAULT_TIMEOUT
        self._timeout = timeout

    def _send_command(self, cmd_list: List[int]) -> None:
        """
        Internal method to check there are correct number of bytes in the input list and sends it to the device.
        :param cmd_list: List[int]. A list of length of four, with "hex" bytes as entries
        """
        if len(cmd_list) != 4:
            raise QMI_InstrumentException(
                "The byte command list must have four byte entries. Input has {}".format(len(cmd_list))
            )

        msg = bytes(cmd_list)
        self._transport.write(msg)

    def _read_reply(self) -> bytes:
        """
        Internal method to query the device for response and to return it to the caller
        :return response: bytes string. Length of four bytes if read succeeded, otherwise empty string.
        """
        response = b""
        try:
            response = self._transport.read(4, timeout=self._timeout)  # We expect to read 4 bytes

        finally:
            return response

    @rpc_method
    def open(self) -> None:
        _logger.info("Opening connection to %s", self._name)
        self._transport.open()
        self._transport.discard_read()
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("Closing connection to %s", self._name)
        super().close()
        self._transport.close()

    @rpc_method
    def get_active_channel(self) -> int:
        """
        Retrieve the current active channel in the switch.
        :return channel_active: int. The active channel number
        """
        self._send_command([0x01, 0x11, 0x00, 0x00])
        reply = self._read_reply()
        channel_active = reply[-1]  # Last byte informs the active channel number

        return channel_active

    @rpc_method
    def set_channel_active(self, channel: int) -> None:
        """
        Set the input channel as the active channel in the switch.
        :param channel: int, the input channel number
        """
        if channel < 1 or channel > self.CHANNELS:
            raise ValueError(
                "Channel number {} not in range of available channels 1-{}".format(channel, self.CHANNELS)
            )

        set_active_channel_list = [0x01, 0x12, 0x00, int(channel)]
        self._send_command(set_active_channel_list)
        # Check the echo that correct channel was set
        reply = self._read_reply()
        try:
            assert reply == bytes(set_active_channel_list)

        except AssertionError:
            raise QMI_InstrumentException("The reply {} does not match request {}. Correct channel not set.")



import typing
import unittest
from unittest.mock import Mock

from qmi.instruments.picotech import _picoscope
from qmi.core.exceptions import QMI_InstrumentException, QMI_UnknownNameException

import tests.instruments.picotech.picosdk_stub
from tests.instruments.picotech.ps3000a_stub import ps3000a
from tests.instruments.picotech.ps4000a_stub import ps4000a


class PicoscopeFind3000AInstrumentsTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self._picoscope = _picoscope.PicoTech_PicoScope
        _picoscope.COMMAND_DICT = {}
        _picoscope._ps = tests.instruments.picotech.ps3000a_stub.ps3000a
        _picoscope._ps.ps3000aEnumerateUnits = Mock(return_value=0)

    def test_01_listing_instruments(self):
        """ This test should run without raising any exceptions and update COMMAND_DICT"""
        self.assertDictEqual(_picoscope.COMMAND_DICT, {})
        expected_stop = {"Stop": ps3000a.ps3000aStop}
        instrus = self._picoscope.list_instruments("3000a")
        self.assertIsInstance(instrus, list)
        self.assertTrue(set(expected_stop).issubset(set(_picoscope.COMMAND_DICT)))

    def test_02_err(self):
        """ Test the error self-check function """
        _picoscope._import_modules("3000a")
        with self.assertRaises(QMI_InstrumentException):
            _picoscope._check_error(1)

    def test_03_import_modules_with_wrong_library_errors(self):
        """ Giving different library name as input that what is the library loaded should raise an exception """
        with self.assertRaises(QMI_UnknownNameException):
            _picoscope._import_modules("4000a")

        self.assertDictEqual(_picoscope.COMMAND_DICT, {})


class PicoscopeFind4000AInstrumentsTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self._picoscope = _picoscope.PicoTech_PicoScope
        _picoscope.COMMAND_DICT = {}
        _picoscope._ps = tests.instruments.picotech.ps4000a_stub.ps4000a
        _picoscope._ps.ps4000aEnumerateUnits = Mock(return_value=0)

    def test_01_listing_instruments(self):
        """ This test should run without raising any exceptions and update COMMAND_DICT"""
        self.assertDictEqual(_picoscope.COMMAND_DICT, {})
        expected_stop = {"Stop": ps4000a.ps4000aStop}
        instrus = self._picoscope.list_instruments("4000a")
        self.assertIsInstance(instrus, list)
        self.assertTrue(set(expected_stop).issubset(set(_picoscope.COMMAND_DICT)))

    def test_02_err(self):
        """ Test the error self-check function """
        _picoscope._import_modules("4000a")
        with self.assertRaises(QMI_InstrumentException):
            _picoscope._check_error(1)

    def test_03_import_modules_with_wrong_library_errors(self):
        """ Giving different library name as input that what is the library loaded should raise an exception """
        with self.assertRaises(QMI_UnknownNameException):
            _picoscope._import_modules("3000a")

        self.assertDictEqual(_picoscope.COMMAND_DICT, {})


class PicotechPicoscopeBaseClassTestCase(unittest.TestCase):
    """Test case for testing the base class functions."""
    def test_not_implemented_methods_raise_exception(self):
        """See that `run_block` and `get_time_resolution` of the base class raise exception when called."""
        with self.assertRaises(NotImplementedError):
            _picoscope.PicoTech_PicoScope(Mock(), "test_run_block", "sn1234").run_block(1, 2, 3)

        with self.assertRaises(NotImplementedError):
            _picoscope.PicoTech_PicoScope(Mock(), "test_get_t_res", "sn1234").get_sampling_interval(1)


if __name__ == "__main__":
    unittest.main()



#! /usr/bin/env python3

"""Test cleanup behaviour of QMI framework."""

import logging
import unittest
import warnings
import weakref
from collections import namedtuple

import gc
import time

import qmi
import qmi.core.exceptions
from qmi.core.context import QMI_Context
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import rpc_method
from qmi.core.task import QMI_Task

# Global variables to inspect the fate of RPC objects.
_instrument_ref = [None]
_instrument_open = [False]
_task_ref = [None]
_task_running = [False]


class SimpleTestInstrument(QMI_Instrument):
    """Dummy instrument used for testing."""

    def __init__(self, context: QMI_Context, name: str) -> None:
        super().__init__(context, name)
        _instrument_ref[0] = weakref.ref(self)

    @rpc_method
    def open(self) -> None:
        super().open()
        _instrument_open[0] = True

    @rpc_method
    def close(self) -> None:
        super().close()
        _instrument_open[0] = False

    @rpc_method
    def hello(self) -> str:
        return "greetings"


class SimpleTestTask(QMI_Task):
    """Dummy task used for testing."""

    Settings = namedtuple("Settings", "x")

    def __init__(self, task_runner, name):
        super().__init__(task_runner, name)
        self.settings = SimpleTestTask.Settings(x=14)
        _task_ref[0] = weakref.ref(self)

    def run(self):
        """Run this task."""

        _task_running[0] = True
        try:

            i = 0
            while True:
                self.sleep(0.2)
                i += 1

        finally:
            _task_running[0] = False


class TestCleanup(unittest.TestCase):

    def setUp(self):
        # Suppress logging.
        logging.getLogger("qmi.core.task").setLevel(logging.ERROR)

    def tearDown(self):
        logging.getLogger("qmi.core.task").setLevel(logging.NOTSET)

    def test_clean_stop(self):
        """Test shutdown with proper cleanup of all created objects."""

        # Start QMI.
        qmi.start("test_context")

        # Create an instrument as RPC object.
        inst = qmi.make_instrument("inst", SimpleTestInstrument)
        inst.open()
        inst.hello()

        # Create a task and start it.
        simple_task = qmi.make_task("simple_task", SimpleTestTask)
        simple_task.start()

        # Wait a bit.
        time.sleep(1.0)

        # Stop task and check that it stopped.
        self.assertTrue(_task_running[0])
        simple_task.stop()
        simple_task.join()
        self.assertFalse(_task_running[0])

        # Remove task and check that the object is destructed.
        self.assertIsInstance(_task_ref[0](), SimpleTestTask)
        qmi.context().remove_rpc_object(simple_task)
        gc.collect()
        self.assertIsNone(_task_ref[0]())

        # Close instrument and check that it got closed.
        inst.hello()
        self.assertTrue(_instrument_open[0])
        inst.close()
        self.assertFalse(_instrument_open[0])

        # Remove instrument and check that the object is destructed.
        self.assertIsInstance(_instrument_ref[0](), SimpleTestInstrument)
        qmi.context().remove_rpc_object(inst)
        gc.collect()
        self.assertIsNone(_instrument_ref[0]())

        # Stop QMI.
        qmi.stop()

    def test_fail_restart(self):
        """Test that a QMI context can not be started a second time."""

        # Create context and start it.
        ctx = QMI_Context("ctx")
        ctx.start()

        # Stop context.
        ctx.stop()

        # Try to restart the context. This should fail.
        with self.assertRaises(qmi.core.exceptions.QMI_UsageException):
            ctx.start()

    def test_unclean_stop(self):
        """Test shutdown without proper cleanup."""

        # Start QMI.
        qmi.start("test_context")

        # Create an instrument as RPC object.
        inst = qmi.make_instrument("inst", SimpleTestInstrument)
        inst.open()
        inst.hello()

        # Create a task and start it.
        simple_task = qmi.make_task("simple_task", SimpleTestTask)
        simple_task.start()

        # Wait a bit.
        time.sleep(1.0)

        # Check references to instrument and task.
        self.assertIsInstance(_instrument_ref[0](), SimpleTestInstrument)
        self.assertIsInstance(_task_ref[0](), SimpleTestTask)
        self.assertTrue(_task_running[0])

        # Stop QMI. This should cleanly remove the remaining objects.
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", ResourceWarning)
            qmi.stop()

        # Check that remaining objects are destructed.
        gc.collect()
        self.assertIsNone(_instrument_ref[0]())
        self.assertIsNone(_task_ref[0]())
        self.assertFalse(_task_running[0])

    def test_stop_handler(self):
        """Test that stop handlers are correctly called."""

        cb1_count = 0
        cb2_count = 0

        def cb1():
            nonlocal cb1_count
            cb1_count += 1

        def cb2():
            nonlocal cb2_count
            cb2_count += 1

        # Start QMI.
        qmi.start("test_context")

        # Setup stop handlers.
        qmi.context().register_stop_handler(cb1)
        qmi.context().register_stop_handler(cb2)

        self.assertEqual(cb1_count, 0)
        self.assertEqual(cb2_count, 0)

        # Stop QMI. This should trigger calls to the stop handlers.
        qmi.stop()

        # Check that the stop handlers were called.
        self.assertEqual(cb1_count, 1)
        self.assertEqual(cb2_count, 1)


if __name__ == "__main__":
    unittest.main()



"""
Instrument driver for the Wavelength Electronics TC Lab temperature controller.
"""

import enum
import logging
import re
import time
from typing import List, NamedTuple, Set, Tuple

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.core.transport import create_transport, list_usbtmc_transports, UsbTmcTransportDescriptorParser


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class AutotuneMode(enum.IntEnum):
    """Autotune modes."""
    MANUAL = 0
    DISTURB_REJECT = 1
    SETPOINT_RESPONSE = 2


class TemperatureControllerCondition(NamedTuple):
    """Condition status register of the temperature controller."""
    current_limit:      bool
    sensor_limit:       bool
    temperature_high:   bool
    temperature_low:    bool
    sensor_shorted:     bool
    sensor_open:        bool
    tec_open:           bool
    in_tolerance:       bool
    output_on:          bool
    laser_shutdown:     bool
    power_on:           bool


class Wavelength_TC_Lab(QMI_Instrument):
    """Instrument driver for the Wavelength Electronics TC Lab temperature controller.

    A subset of the instrument functionality is supported.
    Configuring temperature sensors and sensor parameters is not supported.
    The auxiliary temperature sensor is not supported.
    """

    USB_VENDOR_ID = 0x1a45
    USB_PRODUCT_ID = 0x3101

    # Response timeout for normal commands.
    COMMAND_RESPONSE_TIMEOUT = 2.0

    # The instrument sometimes refuses to report its own USB serial number.
    # As a result, the instrument is not recognized by this driver (about 10%
    # of the attempts to open it).
    # As a workaround we retry opening the instrument a few times.
    OPEN_MAX_RETRY = 5

    @staticmethod
    def list_instruments() -> List[str]:
        """Return a list of QMI transport descriptors for attached instruments."""
        instruments: Set[str] = set()
        for _i in range(Wavelength_TC_Lab.OPEN_MAX_RETRY):
            for transport in list_usbtmc_transports():
                parameters = UsbTmcTransportDescriptorParser.parse_parameter_strings(transport)
                if ((parameters.get("vendorid") == Wavelength_TC_Lab.USB_VENDOR_ID)
                        and (parameters.get("productid") == Wavelength_TC_Lab.USB_PRODUCT_ID)):
                    instruments.add(transport)
        return list(instruments)

    def __init__(self,
                 context: QMI_Context,
                 name: str,
                 transport: str
                 ) -> None:
        """Initialize the instrument driver.

        Parameters:
            name:       Name for this instrument instance.
            transport:  QMI transport descriptor to connect to the instrument.
                        The transport descriptor will typically specify a USBTMC instrument,
                        for example "usbtmc:vendorid=0x1a45:productid=0x3103:serialnr=xxx".
        """
        super().__init__(context, name)
        self._transport = create_transport(transport)
        self._scpi_protocol = ScpiProtocol(self._transport, default_timeout=self.COMMAND_RESPONSE_TIMEOUT)

    def _open_transport(self) -> None:
        """Open transport to the instrument and retry if necessary."""

        # The instrument sometimes refuses to report its own USB serial number.
        # In this case, libusb will not recognize the instrument when we specify
        # it by serial number.
        # As a workaround we retry opening the instrument a few times.
        retry = 0
        while True:
            try:
                self._transport.open()
                return
            except Exception:
                if retry >= self.OPEN_MAX_RETRY:
                    raise
                if retry == 0:
                    _logger.warning("[%s] Failed to open instrument, retrying", self._name)
            time.sleep(0.01)
            retry += 1

    @rpc_method
    def open(self) -> None:
        _logger.info("[%s] Opening connection to instrument", self._name)
        self._check_is_closed()
        self._open_transport()
        try:
            # Discard pending errors.
            self._scpi_protocol.write("*CLS")
            self._scpi_protocol.ask("ERRSTR?")
            # Switch instrument to decimal mode (avoid hex mode).
            self._scpi_protocol.write("RADIX DEC")
        except Exception:
            self._transport.close()
            raise
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("[%s] Closing connection to instrument", self._name)
        super().close()
        self._transport.close()

    def _check_error(self) -> None:
        """Read the instrument error queue and raise an exception if there is an error."""
        resp = self._scpi_protocol.ask("ERRSTR?")
        # When there are no errors, the response is '0,No error'.
        if not re.match(r"^\s*([-+]\s*)?0\s*,", resp):
            # Some error occurred.
            raise QMI_InstrumentException("Instrument returned error: {}".format(resp))

    def _ask_int(self, cmd: str) -> int:
        """Send a query and return an integer response."""
        resp = self._scpi_protocol.ask(cmd)
        try:
            return int(resp)
        except ValueError:
            raise QMI_InstrumentException("Unexpected response to command {!r}: {!r}".format(cmd, resp))

    def _ask_float(self, cmd: str) -> float:
        """Send a query and return a floating point response."""
        resp = self._scpi_protocol.ask(cmd)
        try:
            return float(resp)
        except ValueError:
            raise QMI_InstrumentException("Unexpected response to command {!r}: {!r}".format(cmd, resp))

    @rpc_method
    def reset(self) -> None:
        """Reset instrument to factory default settings."""
        self._scpi_protocol.write("*CLS")  # clear error queue
        self._scpi_protocol.write("*RST")  # factory reset
        self._scpi_protocol.ask("*OPC?")  # wait until reset complete
        self._check_error()

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return QMI_InstrumentIdentification instance."""
        resp = self._scpi_protocol.ask("*IDN?")
        words = resp.rstrip().split(",")
        if len(words) != 4:
            raise QMI_InstrumentException("Unexpected response to *IDN?, got {!r}".format(resp))
        return QMI_InstrumentIdentification(vendor=words[0].strip(),
                                            model=words[1].strip(),
                                            serial=words[2].strip(),
                                            version=words[3].strip())

    @rpc_method
    def get_power_on(self) -> bool:
        """Return True if the device is powered on, False when in standby."""
        return bool(self._ask_int("PWR?"))

    @rpc_method
    def set_power_on(self, power_on: bool) -> None:
        """Switch between power-on and standby mode.

        This corresponds to pressing the "power" button on the front panel.
        While the device is in standby mode, it responds to remote commands
        but many commands will cause the instrument to report error code 992.

        When switching power on, this function returns immediately but it can
        take a few seconds before the instrument is ready to respond to commands.

        Parameters:
            power_on: True to switch to power-on mode, False to switch to standby mode.
        """
        self._scpi_protocol.write("PWR {}".format(1 if power_on else 0))
        self._check_error()

    @rpc_method
    def get_temperature(self) -> float:
        """Read the actual temperature.

        Returns:
            Actual sensor temperature expressed in the selected unit of temperature (see `set_unit()`).
        """
        return self._ask_float("TEC:ACT?")

    @rpc_method
    def get_setpoint(self) -> float:
        """Return the actual temperature setpoint.

        Returns:
            Actual temperature setpoint expressed in the selected unit of temperature (see `set_unit()`).
        """
        return self._ask_float("TEC:SET?")

    @rpc_method
    def set_setpoint(self, setpoint: float) -> None:
        """Change the temperature setpoint.

        It is not possible to set the setpoint outside the temperature limit range (see `set_temperature_limit()`).

        Parameters:
            setpoint: New temperature setpoint expressed in the selected unit of temperature (see `set_unit()`).
        """
        self._scpi_protocol.write("TEC:SET {:.6f}".format(setpoint))
        self._check_error()

    @rpc_method
    def get_unit(self) -> str:
        """Return the selected physical unit of temperature.

        The selected unit is used to report the actual temperature and to set or report the temperature setpoint.

        Returns:
            Unit, either "CELSIUS", "KELVIN", "FAHRENHEIT" or "RAW".
        """
        return self._scpi_protocol.ask("TEC:UNITS?")

    @rpc_method
    def set_unit(self, unit: str) -> None:
        """Change the select physical unit of temperature.

        The selected unit is used to report the actual temperature and to set or report the temperature setpoint.
        The default setting is "CELSIUS".

        Parameters:
            unit: Selected unit, either "CELSIUS", "KELVIN", "FAHRENHEIT" or "RAW".
        """
        unit_str = unit.upper()
        if unit_str not in ("CELSIUS", "KELVIN", "FAHRENHEIT", "RAW"):
            raise ValueError("Unsupported temperature unit")
        self._scpi_protocol.write("TEC:UNITS {}".format(unit_str))
        self._check_error()

    @rpc_method
    def get_tec_current(self) -> float:
        """Read the actual current through the thermoelectric element in Ampere."""
        return self._ask_float("TEC:I?")

    @rpc_method
    def get_tec_voltage(self) -> float:
        """Read the actual voltage on the thermoelectric element in Volt."""
        return self._ask_float("TEC:V?")

    @rpc_method
    def get_output_enabled(self) -> bool:
        """Return True if the controller output is enabled, False if disabled."""
        return bool(self._ask_int("TEC:OUTPUT?"))

    @rpc_method
    def set_output_enabled(self, enable: bool) -> None:
        """Enable or disable the controller output."""
        self._scpi_protocol.write("TEC:OUTPUT {}".format(1 if enable else 0))
        self._check_error()

    @rpc_method
    def get_condition_status(self) -> TemperatureControllerCondition:
        """Read the current condition of the instrument."""
        v = self._ask_int("TEC:COND?")
        return TemperatureControllerCondition(
            current_limit=((v & 1) != 0),
            sensor_limit=((v & 4) != 0),
            temperature_high=((v & 8) != 0),
            temperature_low=((v & 16) != 0),
            sensor_shorted=((v & 32) != 0),
            sensor_open=((v & 64) != 0),
            tec_open=((v & 128) != 0),
            in_tolerance=((v & 512) != 0),
            output_on=((v & 1024) != 0),
            laser_shutdown=((v & 2048) != 0),
            power_on=((v & 32768) != 0))

    @rpc_method
    def get_pid_parameters(self) -> Tuple[float, float, float]:
        """Return the actual PID control parameters.

        Returns:
            Tuple (p, i, d) of PID parameters.
        """
        cmd = "TEC:PID?"
        resp = self._scpi_protocol.ask(cmd)
        fields = resp.split(",")
        if len(fields) != 3:
            raise QMI_InstrumentException("Unexpected response to command {!r}: {!r}".format(cmd, resp))
        try:
            return (float(fields[0]), float(fields[1]), float(fields[2]))
        except ValueError:
            raise QMI_InstrumentException("Unexpected response to command {!r}: {!r}".format(cmd, resp))

    @rpc_method
    def set_pid_parameters(self, p: float, i: float, d: float) -> None:
        """Change the PID control parameters.

        Parameters:
            p: Proportional control parameter, range 0.1 ... 1000.
            i: Integral control parameter, range 0 ... 200.
            d: Derivative control parameter, range 0 ... 100.
        """
        self._scpi_protocol.write("TEC:PID {:.4f},{:.4f},{:.4f}".format(p, i, d))
        self._check_error()

    @rpc_method
    def get_autotune_mode(self) -> AutotuneMode:
        """Read the actual autotune mode."""
        mode = self._ask_int("TEC:AUTOTUNE?")
        return AutotuneMode(mode)

    @rpc_method
    def set_autotune_mode(self, mode: AutotuneMode) -> None:
        """Change the autotune mode.

        In *manual* mode, the PID parameters must be specified explicitly.
        In *disturbance rejection* and *setpoint response* modes, the instrument
        automatically adjusts the PID parameters to match the actual setpoint.

        A one-time characterization scan must be performed before autotune mode
        can be used, see `start_autotune()`.

        Even in autotune mode, the operator can choose to use a derivative
        control parameter (D) or to force it to zero. Call `set_pid_parameters()`
        to set the D parameter either to zero or to any non-zero value,
        the call `set_autotune_mode()` to set the corresponding autotuned parameters.

        Parameters:
            mode: The new autotune mode (one of the `AutotuneMode` constants).
        """
        mode = AutotuneMode(mode)
        self._scpi_protocol.write("TEC:AUTOTUNE {}".format(mode.value))
        self._check_error()

    @rpc_method
    def start_autotune(self) -> None:
        """Start an autotuning characterization scan.

        Before calling this function, first turn off the controller output,
        select one of the autotune modes, and change the setpoint to approximately
        5 degrees away from the ambient temperature. For example::

            instr.set_output_enabled(False)
            instr.set_autotune_mode(AutotuneMode.DISTURB_REJECT)
            instr.set_setpoint(ambient_temp + 5.0)
            instr.start_autotune()

        The characterization scan may take several minutes.
        Once characterization is complete, `get_autotune_valid()` will return True.
        Call `abort_autotune()` to stop the characterization scan and go back to previous settings.
        Do not otherwise interact with the instrument while characterization is in progress.
        """
        self._scpi_protocol.write("TEC:TUNESTART")
        self._check_error()

    @rpc_method
    def abort_autotune(self) -> None:
        """Stop the current autotune characterization scan."""
        self._scpi_protocol.write("TEC:TUNEABORT")
        self._check_error()

    @rpc_method
    def get_autotune_is_valid(self) -> bool:
        """Return True if the instrument has valid autotuning characterization data for the current sensor.

        This variable becomes True after a successful autotuning characterization scan,
        and typically stays True until the sensor type is reconfigured.
        """
        return bool(self._ask_int("TEC:VALID?"))

    @rpc_method
    def get_temperature_limit(self) -> Tuple[float, float]:
        """Return the temperature limits.

        Returns:
            Tuple (temp_low, temp_high) representing the temperature limits
            of the sensor, expressed in the selected unit of temperature
            (see `set_unit()`).
        """
        temp_low = self._ask_float("TEC:LIM:TLO?")
        temp_high = self._ask_float("TEC:LIM:THI?")
        return (temp_low, temp_high)

    @rpc_method
    def set_temperature_limit(self, temp_low: float, temp_high: float) -> None:
        """Configure the temperature limits.

        When the actual temperature exceeds these limits, the controller output is disabled.

        The temperature setpoint is restricted to these limits.

        Parameters:
            temp_low: Low temperature limit, expressed in the selected unit of temperature.
            temp_high: High temperature limit, expressed in th selected unit of temperature.
        """
        self._scpi_protocol.write("TEC:LIM:TLO {:.4f}".format(temp_low))
        self._scpi_protocol.write("TEC:LIM:THI {:.4f}".format(temp_high))
        self._check_error()

    @rpc_method
    def get_tec_current_limit(self) -> Tuple[float, float]:
        """Read the current limits for the thermoelectric element.

        Returns:
            Tuple (lim_pos, lim_neg) representing the maximum positive
            and negative current in Ampere. The maximum negative current
            is reported as a positive number.
        """
        lim_pos = self._ask_float("TEC:LIM:IPOS?")
        lim_neg = self._ask_float("TEC:LIM:INEG?")
        return (lim_pos, lim_neg)

    @rpc_method
    def set_tec_current_limit(self, lim_pos: float, lim_neg: float) -> None:
        """Configure the current limits for the thermoelectric element.

        **Warning:** Setting incorrect current limits can damage the thermoelectric element
        and cause safety hazards.

        The valid range for these settings is from 0 to the maximum current capability of the instrument,
        i.e. 5 A for TC5 LAB, etc.

        To operate a resistive heater, set one of these limits to zero.

        Parameters:
            lim_pos: Maximum positive current in Ampere.
            lim_neg: Maximum negative current in Ampere, represented as a positive number.
        """
        self._scpi_protocol.write("TEC:LIM:IPOS {:.4f}".format(lim_pos))
        self._scpi_protocol.write("TEC:LIM:INEG {:.4f}".format(lim_neg))
        self._check_error()

    @rpc_method
    def get_tec_voltage_limit(self) -> float:
        """Read the actual voltage limit for the thermoelectric element in Volt."""
        return self._ask_float("TEC:VLIM?")

    @rpc_method
    def set_tec_voltage_limit(self, vlim: float) -> None:
        """Set the voltage limit for the thermoelectric element.

        The autotuning process determines an optimal setting for this parameter.

        Parameters:
            vlim: New voltage limit in Volt
                  (valid range 9 ... 18 for TC5 LAB, TC10 LAB, 10 ... 26 for TC15 LAB).
        """
        self._scpi_protocol.write("TEC:VLIM {:.4f}".format(vlim))
        self._check_error()



"""
Siglent Technologies, oscilloscopes and spectrum analyzers.

The qmi.instruments.siglent package provides support for:
- SDS1202X-E oscilloscope
- SSA300X series spectrum analyzers.
"""

from qmi.instruments.siglent.sds1202xe import SDS1202XE, CommHeader, TriggerCondition
from qmi.instruments.siglent.ssa3000x import SSA3000X
# Alternative, QMI naming convention approved names
Siglent_Sds1202xE = SDS1202XE
Siglent_Ssa3000x = SSA3000X



import os
import unittest
from unittest.mock import MagicMock

import qmi
import qmi.utils.context_managers
from qmi.core.transport import QMI_Transport
from qmi.instruments.ozoptics import OzOptics_EPC


class TestEpcDriver(unittest.TestCase):

    def setUp(self) -> None:
        self.transport = MagicMock(spec=QMI_Transport)
        qmi.start("test_ozoptics_epc")
        self.instrument: OzOptics_EPC = qmi.make_instrument('epc_driver', OzOptics_EPC, self.transport)

    def tearDown(self) -> None:
        qmi.stop()

    def test_get_frequency(self):
        # arrange
        self.transport.read_until.side_effect = [b'Frequency(Hz): CH1 007  CH2 017  CH3 037  CH4 071\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            freq_ch1, freq_ch2, freq_ch3, freq_ch4 = self.instrument.get_frequencies()

        # assert
        self.transport.write.assert_called_once_with(b'F?\r\n')
        self.assertEqual(freq_ch1, 7)
        self.assertEqual(freq_ch2, 17)
        self.assertEqual(freq_ch3, 37)
        self.assertEqual(freq_ch4, 71)

    def test_set_frequency(self):
        # arrange
        self.transport.read_until.side_effect = [b'Set CH2 AC Frequency at: 47Hz\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            self.instrument.set_frequency(1, 50)

        # assert
        self.transport.write.assert_called_once_with(b'F1,50\r\n')

    def test_set_frequency_wrong_channel(self):
        # act and assert
        with qmi.utils.context_managers.open_close(self.instrument):
            with self.assertRaises(ValueError):
                self.instrument.set_frequency(0, 50)

        with qmi.utils.context_managers.open_close(self.instrument):
            with self.assertRaises(ValueError):
                self.instrument.set_frequency(5, 50)

    def test_set_frequency_wrong_frequency(self):
        # act and assert
        with qmi.utils.context_managers.open_close(self.instrument):
            with self.assertRaises(ValueError):
                self.instrument.set_frequency(1, 101)

        # act and assert
        with qmi.utils.context_managers.open_close(self.instrument):
            with self.assertRaises(ValueError):
                self.instrument.set_frequency(1, -1)

    def test_set_operating_mode_ac(self):
        # arrange
        self.transport.read_until.side_effect = [b'Set To AC Mode\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            self.instrument.set_operating_mode_ac()

        # assert
        self.transport.write.assert_called_once_with(b'MAC\r\n')

    def test_set_operating_mode_dc(self):
        # arrange
        self.transport.read_until.side_effect = [b'Set To DC Mode\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            self.instrument.set_operating_mode_dc()

        # assert
        self.transport.write.assert_called_once_with(b'MDC\r\n')

    def test_get_operating_mode_ac(self):
        # arrange
        self.transport.read_until.side_effect = [b'Work Mode: AC(Sine)\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            operating_mode_ac = self.instrument.get_operating_mode()

        # assert
        self.transport.write.assert_called_once_with(b'M?\r\n')
        self.assertEqual(operating_mode_ac, 'AC')

    def test_get_operating_mode_dc(self):
        # arrange
        self.transport.read_until.side_effect = [b'Work Mode: DC\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            operating_mode_ac = self.instrument.get_operating_mode()

        # assert
        self.transport.write.assert_called_once_with(b'M?\r\n')
        self.assertEqual(operating_mode_ac, 'DC')

    def test_set_voltage(self):
        # arrange
        self.transport.read_until.side_effect = [b'Set CH3 DC Voltage at: 2500mV\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            self.instrument.set_voltage(3, 2500)

        # assert
        self.transport.write.assert_called_once_with(b'V3,2500\r\n')

    def test_set_voltage_wrong_channel(self):
        # act and assert
        with qmi.utils.context_managers.open_close(self.instrument):
            with self.assertRaises(ValueError):
                self.instrument.set_voltage(0, 2500)

        with qmi.utils.context_managers.open_close(self.instrument):
            with self.assertRaises(ValueError):
                self.instrument.set_voltage(5, 2500)

    def test_set_voltage_wrong_voltage(self):
        # act and assert
        with qmi.utils.context_managers.open_close(self.instrument):
            with self.assertRaises(ValueError):
                self.instrument.set_voltage(3, -5001)

        with qmi.utils.context_managers.open_close(self.instrument):
            with self.assertRaises(ValueError):
                self.instrument.set_voltage(3, 5005)

    def test_set_high(self):
        # arrange
        self.transport.read_until.side_effect = [b'Set CH4 DC Voltage at: 5000mV\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            self.instrument.set_high(4)

        # assert
        self.transport.write.assert_called_once_with(b'VH4\r\n')

    def test_set_low(self):
        # arrange
        self.transport.read_until.side_effect = [b'Set CH4 DC Voltage at: -5000mV\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            self.instrument.set_low(4)

        # assert
        self.transport.write.assert_called_once_with(b'VL4\r\n')

    def test_set_zero(self):
        # arrange
        self.transport.read_until.side_effect = [b'Set CH4 DC Voltage at: 0mV\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            self.instrument.set_zero(4)

        # assert
        self.transport.write.assert_called_once_with(b'VZ4\r\n')

    def test_invalid_channel_voltage_command(self):
        # act and assert
        with qmi.utils.context_managers.open_close(self.instrument):
            with self.assertRaises(ValueError):
                self.instrument.set_high(5)

        with qmi.utils.context_managers.open_close(self.instrument):
            with self.assertRaises(ValueError):
                self.instrument.set_low(0)

        with qmi.utils.context_managers.open_close(self.instrument):
            with self.assertRaises(ValueError):
                self.instrument.set_zero(0)

    def test_get_voltages(self):
        # arrange
        self.transport.read_until.side_effect = [b'Voltage(mV): CH1 +2200 CH2 +5000 CH3 -1000 CH4 -4000\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            voltage_ch1, voltage_ch2, voltage_ch3, voltage_ch4 = self.instrument.get_voltages()

        # assert
        self.transport.write.assert_called_once_with(b'V?\r\n')
        self.assertEqual(voltage_ch1, 2200)
        self.assertEqual(voltage_ch2, 5000)
        self.assertEqual(voltage_ch3, -1000)
        self.assertEqual(voltage_ch4, -4000)

    def test_get_waveform_type_sine(self):
        # arrange
        self.transport.read_until.side_effect = [b'AC Waveform: Sine\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            waveform_type = self.instrument.get_waveform_type()

        # assert
        self.transport.write.assert_called_once_with(b'WF?\r\n')
        self.assertEqual(waveform_type, 'Sine')

    def test_get_waveform_type_triangle(self):
        # arrange
        self.transport.read_until.side_effect = [b'AC Waveform: Triangle\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            waveform_type = self.instrument.get_waveform_type()

        # assert
        self.transport.write.assert_called_once_with(b'WF?\r\n')
        self.assertEqual(waveform_type, 'Triangle')

    def test_save_to_flash(self):
        # arrange
        self.transport.read_until.side_effect = [b'Write Setting to Flash memory\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            self.instrument.save_to_flash()

        # assert
        self.transport.write.assert_called_once_with(b'SAVE\r\n')

    def test_set_waveform_type_sine(self):
        # arrange
        self.transport.read_until.side_effect = [b'Set AC Waveform at: Sine\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            self.instrument.set_waveform_type_sine()

        # assert
        self.transport.write.assert_called_once_with(b'WF1\r\n')

    def test_set_waveform_type_triangle(self):
        # arrange
        self.transport.read_until.side_effect = [b'Set AC Waveform at: Triangle\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            self.instrument.set_waveform_type_triangle()

        # assert
        self.transport.write.assert_called_once_with(b'WF2\r\n')

    def test_enable_dc_in_ac_mode(self):
        # arrange
        self.transport.read_until.side_effect = [b'Enable DC Voltage Output at AC Mode\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            self.instrument.enable_dc_in_ac_mode()

        # assert
        self.transport.write.assert_called_once_with(b'ENVF1\r\n')

    def test_disable_dc_in_ac_mode(self):
        # arrange
        self.transport.read_until.side_effect = [b'Disable DC Voltage Output at AC Mode\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            self.instrument.disable_dc_in_ac_mode()

        # assert
        self.transport.write.assert_called_once_with(b'ENVF0\r\n')

    def test_toggle_channel_ac_dc(self):
        # arrange
        self.transport.read_until.side_effect = [b'Set CH2 output to: Volt (if possible)\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            self.instrument.toggle_channel_ac_dc(2)

        # assert
        self.transport.write.assert_called_once_with(b'VF2\r\n')

    def test_get_ac_dc_status(self):
        # arrange
        self.transport.read_until.side_effect = [b'Hz<->Volt at AC Mode is enabled: CH1-V,CH2-V,CH3-F,CH4-F\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            ch1_mode, ch2_mode, ch3_mode, ch4_mode = self.instrument.get_ac_dc_channel_status()

        # assert
        self.transport.write.assert_called_once_with(b'VF?\r\n')

        self.assertEqual(ch1_mode, 'V')
        self.assertEqual(ch2_mode, 'V')
        self.assertEqual(ch3_mode, 'F')
        self.assertEqual(ch4_mode, 'F')

    def test_is_in_dc_in_ac_mode_enabled(self):
        # arrange
        self.transport.read_until.side_effect = [b'Hz<->Volt at AC Mode is enabled: CH1-V,CH2-V,CH3-F,CH4-F\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            is_dc_in_ac_mode = self.instrument.is_in_dc_in_ac_mode()

        # assert
        self.transport.write.assert_called_once_with(b'VF?\r\n')

        self.assertTrue(is_dc_in_ac_mode)

    def test_is_in_dc_in_ac_mode_disabled(self):
        # arrange
        self.transport.read_until.side_effect = [b'Hz<->Volt at AC Mode is disabled: CH1-V,CH2-V,CH3-F,CH4-F\r\n', b'Done\r\n']

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            is_dc_in_ac_mode = self.instrument.is_in_dc_in_ac_mode()

        # assert
        self.transport.write.assert_called_once_with(b'VF?\r\n')

        self.assertFalse(is_dc_in_ac_mode)



"""Implementation of the QMI_Instrument class.
"""

import logging
import warnings
import typing
from typing import Optional, NamedTuple

from qmi.core.exceptions import QMI_InvalidOperationException
from qmi.core.rpc import QMI_RpcObject, rpc_method


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


# Named tuple to hold result of get_idn() method.
QMI_InstrumentIdentification = NamedTuple(
    'QMI_InstrumentIdentification',
    [('vendor', Optional[str]),
     ('model',  Optional[str]),
     ('serial', Optional[str]),
     ('version', Optional[str])])
#QMI_InstrumentIdentification.__doc__ = """Identification information provided by many types of instruments.
#
#Attributes:
#    vendor: Vendor name.
#    model: Instrument model name or number.
#    serial: Serial number.
#    version: Version or revision number.
#"""


class QMI_Instrument(QMI_RpcObject):
    """Base class for all instrument drivers.

    After an instance of `QMI_Instrument` is created, it must be explicitly
    opened by calling the `open()` method before any interaction with the
    instrument is possible.

    Subclasses implement drivers for specific instrument types.

    All drivers must implement methods `open()` and `close()`
    which initializes resp. tears down the connection to the instrument hardware.

    Subclasses should add methods to implement instrument-specific operations
    (measurements, getting and setting of parameters).

    Driver should implement a method `reset()` when applicable.
    This method returns the instrument to its default settings.

    Drivers should implement a method `get_idn()` when applicable.
    This method returns an instance of `QMI_InstrumentIdentification`.
    """

    @classmethod
    def get_category(cls) -> Optional[str]:
        return "instrument"

    def __init__(self, context: 'qmi.core.context.QMI_Context', name: str) -> None:
        """Initialize the instrument driver instance and store parameters.

        The `__init__()` method should not yet attempt to access the instrument
        hardware. That should be done in the `open()` method.
        """
        super().__init__(context, name)
        self._is_open = False

    @rpc_method
    def __enter__(self) -> "QMI_Instrument":
        """The `__enter__` methods is decorated as `rpc_method` so that `QMI_RpcProxy` can call it when using the
        proxy with a `with` context manager. This method also opens the instrument."""
        self.open()
        return self

    @rpc_method
    def __exit__(self, *args, **kwargs) -> None:
        """The `__exit__` methods is decorated as `rpc_method` so that `QMI_RpcProxy` can call it when using the
        proxy with a `with` context manager. This method also closes the instrument."""
        self.close()

    def release_rpc_object(self) -> None:
        """Give a warning if the instrument is removed while still open."""
        if self._is_open:
            warnings.warn("QMI_Instrument {} removed while still open".format(self._name), ResourceWarning)

    def _check_is_open(self) -> None:
        """Verify that the instrument is open, otherwise raise an exception."""
        if not self._is_open:
            _logger.error("Interaction with closed instrument %s not allowed", self._name)
            raise QMI_InvalidOperationException(
                "Operation not allowed on closed instrument {}".format(self._name))

    def _check_is_closed(self) -> None:
        """Verify that the instrument is closed, otherwise raise an exception."""
        if self._is_open:
            _logger.error("Interaction with open instrument %s not allowed", self._name)
            raise QMI_InvalidOperationException(
                "Operation not allowed on open instrument {}".format(self._name))

    @rpc_method
    def is_open(self) -> bool:
        """Return True if the instrument is open (ready for interaction)."""
        return self._is_open

    @rpc_method
    def open(self) -> None:
        """Connect to the instrument hardware.

        When this method returns, the instrument must be ready for interaction
        via calls to instrument-specific methods.

        Subclasses can extend this method to implement instrument-specific
        initialization. If they do, they should call ``super().open()`` as a last statement.
        """
        self._check_is_closed()
        self._is_open = True

    @rpc_method
    def close(self) -> None:
        """Close the connection to the instrument hardware and release
        associated resources.

        When this method returns, the instrument must not be used again
        unless it is first re-opened by calling the open() method.

        Subclasses can extend this method if they have specific resources to close.
        If they do, they should call ``super().close()`` as a last statement.
        """
        self._check_is_open()
        self._is_open = False


# Imports needed only for static typing.
if typing.TYPE_CHECKING:
    import qmi.core.context
